@@ -7,42 +7,34 @@ use pyo3::{prelude::*, types::PyTuple};
 use rr_util::{
     atr,
     compact_data::U8Set,
-    rearrange_spec::RInnerInts,
-    tensor_util::{IndexError, IndexTensor, TensorIndexSync},
-    tu8v,
-    util::EinsumAxes,
-};
-use rr_util::{
     name::Name,
     opt_einsum::EinsumSpec,
     py_types::{
         einops_repeat, einsum_py, make_diagonal_py, scalar_to_tensor, scalar_to_tensor_py,
-        ExtraPySelfOps, PyEinsumAxes, Tensor, PY_CIRCUIT_ITEMS, PY_UTILS,
+        ExtraPySelfOps, PyEinsumAxes, Tensor, PY_UTILS,
     },
     pycall,
-    rearrange_spec::{OpSize, RearrangeSpec}, // TODO: use new rearrange spec
+    rearrange_spec::{OpSize, RInnerInts, RearrangeSpec},
     sv,
-    tensor_util::{broadcast_shapes, Slice, TensorAxisIndex, TensorIndex, TorchDeviceDtypeOp},
-    util::{
-        cumsum, dict_to_list, {hashmap_collect_except_duplicates, AxisInt},
+    tensor_util::{
+        broadcast_shapes, check_canon_axes, upcast_tensor_device_dtypes, IndexError, IndexTensor,
+        Slice, TensorAxisIndex, TensorIndex, TensorIndexSync, TorchDeviceDtypeOp,
     },
+    tu8v,
+    util::{cumsum, dict_to_list, hashmap_collect_except_duplicates, AxisInt, EinsumAxes},
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
-use smallvec::ToSmallVec;
 use uuid::uuid;
 
 use crate::{
-    check_canon_axes, circuit_node_auto_impl, circuit_node_extra_impl,
+    auto_name::{child_name_with_maybe_paren, children_names_with_maybe_paren, OperatorPriority},
+    circuit_node_auto_impl, circuit_node_extra_impl,
     circuit_node_private::{
         CircuitNodeComputeInfoImpl, CircuitNodeHashItems, CircuitNodeHashWithChildren,
     },
-    circuit_utils::{
-        child_name_with_maybe_paren, children_names_with_maybe_paren, OperatorPriority,
-    },
     new_rc_unwrap,
     prelude::*,
-    upcast_tensor_device_dtypes, CachedCircuitInfo, CircuitNodeSelfOnlyHash, ConstructError,
-    PyCircuitBase, Scalar, Shape,
+    CachedCircuitInfo, CircuitNodeSelfOnlyHash, ConstructError, PyCircuitBase, Scalar, Shape,
 };
 
 /// our Einsum supports diag output
@@ -304,10 +296,6 @@ impl Einsum {
         self.in_axes.iter().cloned().map(PyEinsumAxes).collect()
     }
 
-    fn all_input_circuits(&self) -> Vec<CircuitRc> {
-        self.info().children.clone()
-    }
-
     #[staticmethod]
     pub fn from_nodes_ints(
         nodes: Vec<CircuitRc>,
@@ -618,11 +606,6 @@ impl Add {
     pub fn minus(positive: CircuitRc, negative: CircuitRc, name: Option<Name>) -> Result<Self> {
         positive.sub(negative, name)
     }
-
-    #[getter]
-    pub fn nodes(&self) -> Vec<CircuitRc> {
-        self.info().children.clone()
-    }
 }
 
 circuit_node_extra_impl!(Add, self_hash_default);
@@ -981,12 +964,8 @@ impl CircuitNode for Index {
         assert_eq!(tensors.len(), 1);
         let tensors = upcast_tensor_device_dtypes(tensors);
         Python::with_gil(|py| {
-            PY_CIRCUIT_ITEMS
-                .circ_compiler_util
-                .getattr(py, "IndexUtil")
-                .unwrap()
-                .getattr(py, "apply_dimwise")
-                .unwrap()
+            PY_UTILS
+                .index_apply_dimwise
                 .call(py, (tensors[0].clone(), self.index.clone()), None)
                 .map_err(|err| err.into())
                 .map(|x| x.extract(py).unwrap())
@@ -1025,25 +1004,7 @@ impl Index {
     pub fn slice_edges_to_none(&self) -> Self {
         Index::try_new(
             self.node(),
-            TensorIndex(
-                zip(&self.index.0, &self.node().info().shape)
-                    .map(|(ax, l)| match ax {
-                        TensorAxisIndex::Slice(slice) => TensorAxisIndex::Slice(Slice {
-                            start: if slice.start_u(*l) == 0 {
-                                None
-                            } else {
-                                slice.start
-                            },
-                            stop: if slice.stop_u(*l) == *l {
-                                None
-                            } else {
-                                slice.stop
-                            },
-                        }),
-                        _ => ax.clone(),
-                    })
-                    .collect(),
-            ),
+            self.index.edges_to_none(&self.node().info().shape),
             self.info().name,
         )
         .unwrap()
@@ -1090,11 +1051,7 @@ impl Index {
         let broadcasted_tensor: IndexTensor = pycall!(
             atr!(
                 pycall!(
-                    atr!(
-                        atr!(PY_CIRCUIT_ITEMS.circ_compiler_util, IndexUtil, raw),
-                        just_join_indices_sync,
-                        raw
-                    ),
+                    PY_UTILS.just_join_indices_sync,
                     (
                         tensors_index.clone(),
                         Python::with_gil(|py| PyObject::from(PyTuple::new(
@@ -1351,11 +1308,6 @@ impl Concat {
             .context("failed to unsqueeze for stack (axis invalid?)")?;
         Self::try_new(to_cat, axis, name).context("failed to concat for stack after unsqueeze")
     }
-
-    #[getter]
-    pub fn nodes(&self) -> Vec<CircuitRc> {
-        self.info().children.clone()
-    }
 }
 
 /// Scatter is equivalent to:
@@ -1447,13 +1399,13 @@ impl CircuitNode for Scatter {
 
     fn eval_tensors(&self, tensors: &[Tensor]) -> Result<Tensor> {
         Python::with_gil(|py| {
-            PY_CIRCUIT_ITEMS
-                .circ_compiler_util
-                .getattr(py, "ScatterFn")
-                .unwrap()
-                .call(py, (self.index.clone(), self.info().shape.clone()), None)
-                .unwrap()
-                .call(py, (tensors[0].clone(),), None)
+            PY_UTILS
+                .scatter
+                .call(
+                    py,
+                    (self.index.clone(), self.shape().clone(), tensors[0].clone()),
+                    None,
+                )
                 .map_err(|err| err.into())
                 .map(|x| x.extract(py).unwrap())
         })
@@ -1559,20 +1511,8 @@ impl Conv {
         padding: Vec<(usize, usize)>,
         name: Option<Name>,
     ) -> Result<Self> {
-        let dims = stride.len();
-        if stride.len() != padding.len() {
-            bail!(ConstructError::ConvStridePaddingNotFull {
-                stride: stride.len(),
-                padding: padding.len(),
-            });
-        }
-        if dims > 3 {
-            bail!(ConstructError::NotSupportedYet {
-                string: "Conv over more than 3 dims not supported".into(),
-            });
-        }
-
-        let shape = Self::try_compute_shape(&input, &filter, &stride, &padding)?;
+        let shape =
+            rr_util::conv_shape::conv_shape(input.shape(), filter.shape(), &stride, &padding)?;
 
         let out = Self {
             stride,
@@ -1582,77 +1522,6 @@ impl Conv {
         // todo check not too many axes / oob
         out.initial_init_info()
     }
-
-    fn try_compute_shape(
-        input: &CircuitRc,
-        filter: &CircuitRc,
-        stride: &[usize],
-        padding: &[(usize, usize)],
-    ) -> Result<Shape> {
-        let dims = stride.len();
-        let input_batch_shape =
-            if let Some(input_batch_rank) = input.info().rank().checked_sub(dims + 1) {
-                Ok::<Shape, ConstructError>(input.info().shape[..input_batch_rank].to_smallvec())
-            } else {
-                bail!(ConstructError::ConvInputWrongShape {});
-            }?;
-
-        let filter_batch_shape =
-            if let Some(filter_batch_rank) = filter.info().rank().checked_sub(dims + 2) {
-                Ok::<Shape, ConstructError>(input.info().shape[..filter_batch_rank].to_smallvec())
-            } else {
-                bail!(ConstructError::ConvFilterWrongShape {});
-            }?;
-        let output_batch_shape =
-            broadcast_shapes(&vec![&input_batch_shape[..], &filter_batch_shape[..]])
-                .context("input and filter batch shapes dont broadcast")?;
-
-        if !filter_batch_shape.is_empty() {
-            bail!(ConstructError::NotSupportedYet {
-                string: format!(
-                    "Batch dims on filter aren't supported yet, found {:?}",
-                    filter_batch_shape
-                ),
-            });
-        }
-        let out_channels = filter.info().shape[filter_batch_shape.len()];
-        let in_channels = filter.info().shape[filter.info().rank() - 1];
-        if in_channels != input.info().shape[input.info().rank() - 1] {
-            bail!(ConstructError::ConvInputFilterDifferentNumInputChannels {
-                filter: in_channels,
-                input: input.info().shape[input.info().rank() - 1],
-            });
-        }
-        let input_conv_shape =
-            &input.info().shape[input_batch_shape.len()..input_batch_shape.len() + dims];
-        let filter_kernel_shape =
-            &filter.info().shape[filter.info().rank() - dims - 1..filter.info().rank() - 1];
-        let input_conv_shape_padded_minus_edges: Shape =
-            zip(input_conv_shape, zip(padding, filter_kernel_shape))
-                .map(|(input_l, (padding, kernel_l))| {
-                    input_l - (kernel_l - 1) + padding.0 + padding.1
-                })
-                .collect();
-        let new_conv_shape: Shape = zip(&input_conv_shape_padded_minus_edges, stride)
-            .map(|(i, s)| {
-                if i % s != 0 {
-                    Err(ConstructError::ConvStrideMustDivide {
-                        shape: input_conv_shape_padded_minus_edges.clone(),
-                        stride: stride.to_vec(),
-                    })
-                    .context("hi")
-                } else {
-                    Ok(i / s)
-                }
-            })
-            .collect::<Result<Shape>>()?;
-
-        Ok(output_batch_shape
-            .into_iter()
-            .chain(new_conv_shape)
-            .chain(std::iter::once(out_channels))
-            .collect())
-    }
 }
 
 circuit_node_extra_impl!(Conv, self_hash_default);
