@@ -1,4 +1,5 @@
 use anyhow::Result;
+use itertools::izip;
 use macro_rules_attribute::apply;
 use pyo3::prelude::*;
 use rr_util::{
@@ -10,9 +11,9 @@ use rr_util::{
 };
 
 use crate::{
+    auto_name::{child_name_with_maybe_paren, OperatorPriority},
     circuit_node_auto_impl, circuit_node_extra_impl,
     circuit_node_private::{CircuitNodeComputeInfoImpl, CircuitNodeHashItems},
-    circuit_utils::{child_name_with_maybe_paren, OperatorPriority},
     new_rc_unwrap, CachedCircuitInfo, CircuitFlags, CircuitNode, CircuitNodeAutoName, CircuitRc,
     PyCircuitBase, TensorEvalError,
 };
@@ -67,7 +68,10 @@ impl CircuitNode for SetSymbolicShape {
     circuit_node_auto_impl!("25332a58-39d6-443d-b225-65c810a09aa4");
 
     fn child_axis_map(&self) -> Vec<Vec<Option<usize>>> {
-        vec![(0..self.rank()).map(Some).collect()]
+        vec![izip!(self.shape(), self.node().shape())
+            .enumerate()
+            .map(|(i, (a, b))| if a == b { Some(i) } else { None })
+            .collect()]
     }
 
     fn _replace_children(&self, children: Vec<CircuitRc>) -> Result<Self> {
