@@ -8,9 +8,9 @@ use anyhow::{bail, Context, Error, Result};
 use bit_iter::BitIter;
 use circuit_base::{
     computational_node::EinsumArgs, prelude::*, Add, CircuitType, Concat, DiscreteVar, Einsum,
-    GeneralFunction, Index, Module, Rearrange, Scalar, Scatter, Symbol, Tag,
+    GeneralFunction, Index, Module, ModuleSpec, Rearrange, Scalar, Scatter, Symbol, Tag,
 };
-use itertools::Itertools;
+use itertools::{izip, Itertools};
 use macro_rules_attribute::apply;
 use pyo3::{exceptions::PyValueError, prelude::*};
 use rr_util::{
@@ -18,10 +18,10 @@ use rr_util::{
     filter_by_variant,
     name::Name,
     opt_einsum::optimize_einsum_spec_cached,
-    py_types::PY_CIRCUIT_ITEMS,
     python_error_exception,
     rearrange_spec::{shape_to_op_shape, OpSize, RInts, RearrangeSpec},
     sv,
+    symbolic_size::SymbolicSizeProduct,
     tensor_util::{
         check_canon_idxs, compose, IndexTensor, Shape, Slice, TensorAxisIndex, TensorIndex, USlice,
     },
@@ -1285,6 +1285,7 @@ pub fn push_down_index_raw(
     allow_partial_pushdown: bool,
     call_on_sub: &mut dyn FnMut(usize, Index) -> Result<CircuitRc>,
     suffix: Option<String>,
+    into_module_spec: bool,
 ) -> Result<CircuitRc> {
     let out_name = node.node().info().name.map(|n_s| {
         suffix
@@ -1600,26 +1601,13 @@ pub fn push_down_index_raw(
             }
         }
         Circuit::Rearrange(inner) => {
-            // doing this in python instead of porting to rust bc it has a lot of torch ops so likely wouldn't be faster in rust
-            let py_out = Python::with_gil(|py| {
-                PY_CIRCUIT_ITEMS
-                    .circ_compiler_util
-                    .getattr(py, "index_before_rearrange_rust")
-                    .unwrap()
-                    .call(
-                        py,
-                        (
-                            node.index.clone(),
-                            inner.spec.clone(),
-                            inner.node().info().shape.clone(),
-                            node.info().device_dtype.clone().unwrap_or_defaults().device,
-                        ),
-                        None,
-                    )
-                    .unwrap()
-                    .extract(py)
-            })?;
-            let (index, spec): (TensorIndex, RearrangeSpec) = if let Some(x) = py_out {
+            let (index, spec): (TensorIndex, RearrangeSpec) = if let Some(x) = inner
+                .conform_to_input_shape_spec()
+                .canonicalize(true)
+                .index_before(
+                    node.index.canonicalize(&node.node().shape()),
+                    node.info().device_dtype.clone().unwrap_or_defaults().device,
+                ) {
                 x
             } else {
                 bail!(PushDownIndexError::RearrangeNotPossible {
@@ -1701,37 +1689,110 @@ pub fn push_down_index_raw(
         ),
         Circuit::Scalar(_) => index_merge_scalar(node).unwrap(),
         Circuit::Module(m) => {
-            // doesn't push anything into spec,
+            // TODO: could just use child_axis_map always here? also could write push_down_index default case using child axis map?
+
+            // doesn't push anything into spec, pushes into args when batching over them,
+            // or if spec is batchable to input according to child_axis_map
             let batch_shape = m.aligned_batch_shape();
             let rank_to_pass = batch_shape.len();
 
-            let new_nodes = m
-                .args()
-                .zip(&m.spec.arg_specs)
+            let mut index_non_batch_remaining = node.index.0[rank_to_pass..].to_vec();
+            let index_non_batch: Vec<TensorIndex> = if izip!(
+                &node.index.0[rank_to_pass..],
+                &m.shape()[rank_to_pass..]
+            )
+            .any(|(ix, l)| !ix.is_identity(*l))
+            {
+                let map = m.child_axis_map_inputs().unwrap();
+                izip!(map, m.args_slice(), &m.spec.arg_specs)
+                        .map(|(m2, input_node, arg_spec)| {
+                            let mut ix = vec![TensorAxisIndex::IDENT; arg_spec.symbol.rank()];
+                            let this_num_batch = input_node.rank() - arg_spec.symbol.rank();
+                            m2.iter().enumerate().for_each(|(child_axis, p)| {
+                                if let Some(parent_axis) = p.clone() &&
+                                        parent_axis >= rank_to_pass &&
+                                        arg_spec.expandable &&
+                                        (SymbolicSizeProduct::has_symbolic(arg_spec.symbol.shape()[child_axis - this_num_batch])
+                                         || !arg_spec.ban_non_symbolic_size_expand)
+                                        {
+                                        let parent_axis_c = parent_axis - rank_to_pass;
+                                        if !matches!(
+                                            index_non_batch_remaining[parent_axis_c],
+                                            TensorAxisIndex::Single(_)
+                                        ) {
+                                            index_non_batch_remaining[parent_axis_c] =
+                                                TensorAxisIndex::IDENT;
+                                            ix[child_axis - this_num_batch] = node.index.0[parent_axis].clone();
+                                        }
+                                    };
+                            });
+
+                            TensorIndex(ix)
+                        })
+                        .collect()
+            } else {
+                m.spec
+                    .arg_specs
+                    .iter()
+                    .map(|arg_spec| {
+                        TensorIndex(vec![TensorAxisIndex::IDENT; arg_spec.symbol.rank()])
+                    })
+                    .collect()
+            };
+
+            let new_nodes = izip!(m.args(), &m.spec.arg_specs, index_non_batch)
                 .enumerate()
-                .map(|(i, (input_node, arg_spec))| {
+                .map(|(i, (input_node, arg_spec, ix_non_batch))| {
                     let this_num_batch = input_node.rank() - arg_spec.symbol.rank();
                     assert!(this_num_batch <= rank_to_pass);
                     let start = rank_to_pass - this_num_batch;
-                    let index_passed = TensorIndex(
-                        node.index.0[start..rank_to_pass]
-                            .iter()
-                            .cloned()
-                            .chain(vec![TensorAxisIndex::IDENT; arg_spec.symbol.rank()])
-                            .collect(),
-                    );
+                    let index_passed =
+                        TensorIndex([&node.index.0[start..rank_to_pass], &ix_non_batch.0].concat());
                     on_sub(i + 1, input_node.clone(), index_passed, None)
                 })
                 .collect::<Result<_>>()?;
-            let new_mod = Module::nrc(new_nodes, m.spec.clone(), out_name.clone());
-            handle_non_batchable_remaining(rank_to_pass, new_mod).map_err(
-                |(_, top_axis_indices)| PushDownIndexError::ModuleSomeAxesNotPossible {
-                    index_node: node.clone(),
-                    num_non_batch_out: m.rank() - rank_to_pass,
-                    inner_rank: m.rank(),
-                    top_axis_indices,
-                },
-            )?
+
+            let idxnb = TensorIndex(index_non_batch_remaining);
+
+            if idxnb.is_identity(&m.shape()[rank_to_pass..]) {
+                Module::nrc(new_nodes, m.spec.clone(), out_name)
+            } else {
+                if into_module_spec {
+                    Module::nrc(
+                        new_nodes,
+                        ModuleSpec::new(
+                            on_sub(0, m.spec.circuit.clone(), idxnb, None)?,
+                            m.spec.arg_specs.clone(),
+                            false,
+                            false,
+                        )?,
+                        out_name,
+                    )
+                } else {
+                    let rank_done_now = rank_to_pass
+                        - (0..rank_to_pass)
+                            .filter(|i| matches!(node.index.0[*i], TensorAxisIndex::Single(_)))
+                            .count();
+                    let remaining_index = TensorIndex(
+                        [&vec![TensorAxisIndex::IDENT; rank_done_now], &idxnb.0[..]].concat(),
+                    );
+                    let index = Index::new(
+                        Module::nrc(new_nodes, m.spec.clone(), out_name),
+                        remaining_index,
+                        idx_partial_name(),
+                    );
+                    if allow_partial_pushdown {
+                        index.rc()
+                    } else {
+                        bail!(PushDownIndexError::ModuleSomeAxesNotPossible {
+                            index_node: node.clone(),
+                            num_non_batch_out: m.rank() - rank_to_pass,
+                            inner_rank: m.rank(),
+                            remaining_index: idxnb,
+                        })
+                    }
+                }
+            }
         }
 
         // Circuit::Cumulant(cum) => Cumulant::nrc(
@@ -1762,11 +1823,12 @@ pub fn push_down_index_once(
         allow_partial_pushdown.unwrap_or(false),
         &mut |_, x| Ok(x.rc()),
         suffix,
+        false,
     )
 }
 
 pub fn push_down_index_op(node: &Index) -> Option<CircuitRc> {
-    push_down_index_raw(node, true, &mut |_, x| Ok(x.rc()), None).ok()
+    push_down_index_raw(node, true, &mut |_, x| Ok(x.rc()), None, false).ok()
 }
 
 #[pyfunction]
@@ -2078,12 +2140,12 @@ pub enum PushDownIndexError {
         inner_rank: usize,
         top_axis_indices: Vec<TensorAxisIndex>,
     },
-    #[error("{MODULE_AXIS_NO_MSG}\nindex_node={index_node:?} num_non_batch_out={num_non_batch_out} inner_rank={inner_rank} top_axis_indices={top_axis_indices:?} ({e_name})")]
+    #[error("{MODULE_AXIS_NO_MSG}\nindex_node={index_node:?} num_non_batch_out={num_non_batch_out} inner_rank={inner_rank} remaining_index={remaining_index:?} ({e_name})")]
     ModuleSomeAxesNotPossible {
         index_node: Index,
         num_non_batch_out: usize,
         inner_rank: usize,
-        top_axis_indices: Vec<TensorAxisIndex>,
+        remaining_index: TensorIndex,
     },
     // TODO: improve errors as needed
     #[error("index_node={index_node:?} ({e_name})")]
