@@ -43,9 +43,16 @@ pub fn batch_to_concat(
         bail!(BatchError::PartialBatchNotAllowed {});
     }
     let circuit = deep_map_unwrap(circuit, |x| {
-        x.as_index()
-            .map(|x| Index::slice_edges_to_none(x).rc())
-            .unwrap_or(x.clone())
+        match &**x {
+            Circuit::Index(index) => Index::slice_edges_to_none(&index).rc(),
+            // Circuit::Scatter(scatter) => Scatter::nrc(
+            //     scatter.node(),
+            //     scatter.index.edges_to_none(&scatter.info().shape),
+            //     scatter.info().shape.clone(),
+            //     scatter.info().name,
+            // ),
+            _ => x.clone(),
+        }
     });
 
     let (name, circuit, leaves_axis, leaves_non_axis) = get_axis_leaves(circuit, axis)?;
