@@ -28,6 +28,7 @@ use crate::{
     compiler_strip::strip_names_and_tags,
     deep_rewrite::{
         compiler_simp, deep_heuristic_nest_adds, deep_optimize_einsums, deep_push_down_index_raw,
+        SimpFnSubset,
     },
     module_rewrite::deep_module_remove_unused_inputs,
     scheduled_execution::{
@@ -41,7 +42,7 @@ use crate::{
 // difficult
 
 #[pyclass]
-#[derive(Debug, Clone, Copy)]
+#[derive(Debug, Clone)]
 pub struct OptimizationSettings {
     #[pyo3(get, set)]
     pub verbose: usize,
@@ -81,6 +82,13 @@ pub struct OptimizationSettings {
     pub optimization_parallelism: Option<usize>,
     #[pyo3(get, set)]
     pub keep_all_names: bool,
+    #[pyo3(get, set)]
+    pub push_down_index: bool,
+    #[pyo3(get, set)]
+    pub deep_maybe_distribute: bool,
+    #[pyo3(get, set)]
+    /// plausibly this should instead by SimpFnSubset which defaults to compiler subset
+    pub simp_fn_subset: SimpFnSubset,
 }
 
 impl Default for OptimizationSettings {
@@ -105,6 +113,9 @@ impl Default for OptimizationSettings {
             save_circ: false,
             optimization_parallelism: Some(8),
             keep_all_names: false,
+            push_down_index: true,
+            deep_maybe_distribute: true,
+            simp_fn_subset: SimpFnSubset::compiler_default(),
         }
     }
 }
@@ -131,7 +142,10 @@ impl OptimizationSettings {
         log_slow_einsums = OptimizationSettings::default().log_slow_einsums,
         save_circ = OptimizationSettings::default().save_circ,
         optimization_parallelism = None,
-        keep_all_names = OptimizationSettings::default().keep_all_names
+        keep_all_names = OptimizationSettings::default().keep_all_names,
+        push_down_index = OptimizationSettings::default().push_down_index,
+        deep_maybe_distribute = OptimizationSettings::default().deep_maybe_distribute,
+        simp_fn_subset = OptimizationSettings::default().simp_fn_subset
     ))]
     fn new(
         verbose: usize,
@@ -153,6 +167,9 @@ impl OptimizationSettings {
         save_circ: bool,
         optimization_parallelism: Option<usize>,
         keep_all_names: bool,
+        push_down_index: bool,
+        deep_maybe_distribute: bool,
+        simp_fn_subset: SimpFnSubset,
     ) -> Self {
         Self {
             verbose,
@@ -174,6 +191,9 @@ impl OptimizationSettings {
             save_circ,
             optimization_parallelism,
             keep_all_names,
+            push_down_index,
+            deep_maybe_distribute,
+            simp_fn_subset,
         }
     }
 }
@@ -284,14 +304,18 @@ pub fn optimize_circuit(
     );
     // originally tried push_down_index in compiler_simp, but that produced worse circuits
     // for unknown reasons, maybe i'll investigate
-    let circuit = timed!(
-        apply_fn_until_same(&circuit, |x| deep_push_down_index_raw(
-            compiler_simp(x.clone(), context),
-            None
-        )),
-        10,
-        print_timings
-    );
+    let circuit = if context.settings.push_down_index {
+        timed!(
+            apply_fn_until_same(&circuit, |x| deep_push_down_index_raw(
+                compiler_simp(x.clone(), context),
+                None
+            )),
+            10,
+            print_timings
+        )
+    } else {
+        circuit
+    };
     let circuit = timed!(compiler_simp(circuit, context), 10, print_timings);
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
     if context.settings.verbose >= 3 {
@@ -301,14 +325,18 @@ pub fn optimize_circuit(
             .unwrap();
         print_circuit_stats(&*circuit);
     }
-    let circuit = timed!(
-        apply_fn_until_same(&circuit, |x| {
-            let distributed = deep_maybe_distribute(x.clone(), context);
-            compiler_simp(distributed, context)
-        }),
-        10,
-        print_timings
-    );
+    let circuit = if context.settings.deep_maybe_distribute {
+        timed!(
+            apply_fn_until_same(&circuit, |x| {
+                let distributed = deep_maybe_distribute(x.clone(), context);
+                compiler_simp(distributed, context)
+            }),
+            10,
+            print_timings
+        )
+    } else {
+        circuit
+    };
 
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
     if context.settings.verbose >= 3 {
@@ -340,7 +368,7 @@ pub fn optimize_to_schedule(
     circuit: CircuitRc,
     settings: OptimizationSettings,
 ) -> Result<Schedule> {
-    let context = &mut OptimizationContext::new_settings(settings);
+    let context = &mut OptimizationContext::new_settings(settings.clone());
     let verbose = settings.verbose;
     if verbose > 0 {
         python_println!("Optimizing verbose {}", verbose)
@@ -394,7 +422,7 @@ pub fn optimize_and_evaluate_many(
         }
     }
 
-    let schedule = optimize_to_schedule_many(circuits, settings)?;
+    let schedule = optimize_to_schedule_many(circuits, settings.clone())?;
     schedule.evaluate_many(settings)
 }
 
