@@ -2,10 +2,10 @@ use std::simd::{u64x8, u8x8, SimdPartialEq};
 
 use anyhow::{bail, Result};
 use circuit_base::{
-    circuit_utils::deep_replace_parents_fix, deep_map_op_context, deep_map_preorder_unwrap,
-    deep_map_unwrap, get_compatible_dtype, on_circuit_names, prelude::*, visit_circuit_unwrap, Add,
-    Index,
+    deep_map_op_context, deep_map_preorder_unwrap, deep_map_unwrap, get_compatible_dtype,
+    on_circuit_names, prelude::*, visit_circuit_unwrap, Add, Index,
 };
+use circuit_utils::deep_replace_parents_fix;
 use itertools::{izip, Itertools};
 use macro_rules_attribute::apply;
 use num_bigint::BigUint;
