@@ -6,13 +6,14 @@ use std::{
 use anyhow::{anyhow, bail, Context, Result};
 use circuit_base::{
     cached_circuit_properties::max_non_leaf_size,
-    circuit_utils::{hash_to_node_non_free, is_definitely_view_on_child, toposort_circuit},
+    circuit_utils::toposort_circuit,
     flat_concat,
     module::{any_children_with_symbolic_sizes, substitute_all_modules},
     prelude::*,
     print::{oom_fmt, PrintOptions},
     visit_circuit_non_free, Array, IrreducibleNode, Leaf, ModuleSpec, Scalar, Symbol,
 };
+use circuit_utils::{hash_to_node_non_free, is_definitely_view_on_child};
 use itertools::Itertools;
 use macro_rules_attribute::apply;
 use miniserde::json;
