@@ -11,9 +11,9 @@ use circuit_base::{
     opaque_iterative_matcher::{
         Update as BaseUpdate, UpdatedIterativeMatcher as BaseUpdatedIterativeMatcher,
     },
-    set_of_circuits::SetOfCircuitIdentities,
     CircuitNode, CircuitNodeUnion, CircuitRc, CircuitType,
 };
+use circuit_utils::SetOfCircuitIdentities;
 use macro_rules_attribute::apply;
 use pyo3::{
     exceptions::{PyTypeError, PyValueError},
@@ -672,6 +672,22 @@ impl IterativeMatcher {
     // TODO: more rust niceness funcs like the py ones!
 }
 
+impl IterativeMatcherRc {
+    // -> (found, all_finished, children matchers)
+    pub fn match_iterate_continue(
+        self,
+        circuit: CircuitRc,
+    ) -> Result<(bool, bool, Vec<IterativeMatcherRc>)> {
+        let num_children = circuit.num_children();
+        let (updated, found) = self.match_iterate(circuit)?.unwrap_or_same(self);
+        Ok((
+            found,
+            updated.all_finished(),
+            updated.per_child_with_term(num_children),
+        ))
+    }
+}
+
 #[pymethods]
 impl IterativeMatcher {
     #[new]
