@@ -1,6 +1,6 @@
 use anyhow::Result;
 use circuit_base::{CircuitNode, CircuitRc, Symbol};
-use circuit_rewrites::circuit_manipulation::replace_nodes_py;
+use circuit_utils::replace_nodes_py;
 use macro_rules_attribute::apply;
 use pyo3::prelude::*;
 use rr_util::{cached_method, caching::FastUnboundedCache, name::Name, pycall, util::HashBytes};
