@@ -6,9 +6,9 @@ use std::{
 use anyhow::{anyhow, bail, Result};
 use circuit_base::{
     expand_node::{expand_node, ReplaceMapRc},
-    set_of_circuits::SetOfCircuitIdentities,
     CircuitNode, CircuitRc,
 };
+use circuit_utils::SetOfCircuitIdentities;
 use macro_rules_attribute::apply;
 use pyo3::prelude::*;
 use rr_util::{
