@@ -8,10 +8,6 @@ use circuit_base::{
     visit_circuit, Add, Array, CircResult, Circuit, CircuitNode, CircuitRc, CircuitType, Cumulant,
     DiscreteVar, Einsum, GeneralFunction, Index, Scalar,
 };
-use circuit_rewrites::{
-    circuit_optimizer::OptimizationSettings,
-    generalfunction_rewrite::generalfunction_gen_index_const_to_index,
-};
 use itertools::Itertools;
 use macro_rules_attribute::apply;
 use pyo3::{exceptions::PyValueError, prelude::*};
@@ -114,8 +110,6 @@ pub struct RandomSampleSpec {
     #[pyo3(get)]
     replace: bool,
     #[pyo3(get, set)]
-    probs_and_group_evaluation_settings: OptimizationSettings,
-    #[pyo3(get, set)]
     seeder: Seeder,
     #[pyo3(get, set)]
     simplify: bool,
@@ -188,14 +182,12 @@ impl RandomSampleSpec {
     #[pyo3(signature=(
         shape = RandomSampleSpec::default().shape,
         replace = RandomSampleSpec::default().replace,
-        probs_and_group_evaluation_settings = RandomSampleSpec::default().probs_and_group_evaluation_settings,
         simplify = RandomSampleSpec::default().simplify,
         seeder = None
     ))]
     fn new(
         shape: Shape,
         replace: bool,
-        probs_and_group_evaluation_settings: OptimizationSettings,
         simplify: bool,
         seeder: Option<Seeder>,
     ) -> PyClassInitializer<Self> {
@@ -204,7 +196,6 @@ impl RandomSampleSpec {
             shape,
             replace,
             simplify,
-            probs_and_group_evaluation_settings,
             seeder,
             ..Default::default()
         }
@@ -252,7 +243,6 @@ impl Default for RandomSampleSpec {
             shape: Sv::new(),
             replace: true,
             simplify: true,
-            probs_and_group_evaluation_settings: Default::default(),
             seeder: default_hash_seeder(Some(0)), // NOTE: this is deterministic!
         }
     }
@@ -260,37 +250,40 @@ impl Default for RandomSampleSpec {
 
 impl SampleSpecTrait for RandomSampleSpec {
     fn sample_var(&self, var: &DiscreteVar) -> Result<CircuitRc> {
+        let name = var.info().name.map(|n| format!("{n} sampled").into());
+        let values = var.values().clone();
+        let values_ndim = values.ndim();
+
         let index = self.sample_impl(var.probs_and_group().clone())?;
-        let index = if self.simplify
+        if self.simplify
             && index.info().is_explicitly_computable()
             && var.probs_and_group().is_tag()
             && let Circuit::Tag(c) = &***var.probs_and_group()
             && c.node().is_leaf_constant()
+            && var.values().info().device_dtype.device.is_some()
+            && let Ok(index) = evaluate(index.clone())
+            && index.ndim() == 1
         {
-            evaluate(index.clone())
-                .and_then(|x| Ok(Array::try_new(
-                    TorchDeviceDtypeOp{device: var.values().info().device_dtype.device, dtype: None}.cast_tensor(x),
-                    index.info().name)?.rc(),)).unwrap_or(index)
-        } else {
-            index
+                let index = TorchDeviceDtypeOp{device: var.values().info().device_dtype.device, dtype: None}.cast_tensor(index);
+                return Ok(Index::try_new(
+                    values,
+                    TensorIndex::new_single(TensorAxisIndex::Tensor(index.try_into().unwrap()), 0, values_ndim),
+                    name
+                )?.rc())
         };
         let batch_x = true;
         let batch_index = false;
-        let res = GeneralFunction::gen_index(
-            var.values().clone(),
+        Ok(GeneralFunction::gen_index(
+            values,
             index,
-            -(var.values().ndim() as i64),
+            -(values_ndim as i64),
             batch_x,
             batch_index,
             true,
-            var.info().name.map(|n| format!("{n} sampled").into()),
+            name,
         )
-        .unwrap();
-        Ok(if self.simplify {
-            generalfunction_gen_index_const_to_index(&res).unwrap_or(res.rc())
-        } else {
-            res.rc()
-        })
+        .unwrap()
+        .rc())
     }
     fn get_empirical_weights(&self) -> CircuitRc {
         Scalar::new(
