@@ -1,7 +1,7 @@
 use anyhow::Result;
 use circuit_base::{prelude::*, CircuitType, Index};
 use circuit_rewrites::algebraic_rewrite::{index_elim_identity, push_down_index_raw};
-use get_update_node::{new_traversal, IterativeMatcherRc, Matcher};
+use get_update_node::{new_traversal, IterativeMatcherRc, Matcher, MatcherData};
 use macro_rules_attribute::apply;
 use pyo3::prelude::*;
 use rr_util::{cached_lambda, util::HashBytes};
@@ -20,6 +20,7 @@ pub fn default_index_traversal() -> IterativeMatcherRc {
 #[pyo3(signature=(
     index,
     traversal = default_index_traversal(),
+    into_module_spec = MatcherData::Always(false).into(),
     suffix = None,
     allow_partial_pushdown = false,
     elim_identity = true,
@@ -27,37 +28,47 @@ pub fn default_index_traversal() -> IterativeMatcherRc {
 pub fn push_down_index(
     index: Index,
     traversal: IterativeMatcherRc,
+    into_module_spec: IterativeMatcherRc,
     suffix: Option<String>,
     allow_partial_pushdown: bool,
     elim_identity: bool,
 ) -> Result<CircuitRc> {
     #[apply(cached_lambda)]
-    #[key((index.info().hash, traversal.clone()), (HashBytes, IterativeMatcherRc))]
+    #[key((index.info().hash, traversal.clone(), into_module_spec.clone()), (HashBytes, IterativeMatcherRc, IterativeMatcherRc))]
     #[use_try]
-    fn push_down_rec(index: Index, traversal: IterativeMatcherRc) -> Result<CircuitRc> {
+    fn push_down_rec(
+        index: Index,
+        traversal: IterativeMatcherRc,
+        into_module_spec: IterativeMatcherRc,
+    ) -> Result<CircuitRc> {
         if elim_identity {
             if let Some(removed_node) = index_elim_identity(&index) {
                 return Ok(removed_node);
             }
         }
-        let updated = traversal
-            .match_iterate(index.node().clone())?
-            .unwrap_or_same(traversal)
-            .0;
+        let (_, finished, traversal_per_child) =
+            traversal.match_iterate_continue(index.node().clone())?;
+        let (into_module_spec_here, _, into_module_spec_per_child) =
+            into_module_spec.match_iterate_continue(index.node().clone())?;
 
-        if updated.all_finished() {
+        if finished {
             return Ok(index.rc());
         }
 
-        let traversal_per_child = updated.per_child_with_term(index.node().num_children());
-
         push_down_index_raw(
             &index,
             allow_partial_pushdown,
-            &mut |i, new_index| push_down_rec(new_index, traversal_per_child[i].clone()),
+            &mut |i, new_index| {
+                push_down_rec(
+                    new_index,
+                    traversal_per_child[i].clone(),
+                    into_module_spec_per_child[i].clone(),
+                )
+            },
             suffix.clone(),
+            into_module_spec_here,
         )
     }
 
-    push_down_rec(index, traversal)
+    push_down_rec(index, traversal, into_module_spec)
 }
