@@ -6,7 +6,7 @@ use std::{
 
 use anyhow::{anyhow, bail, Context, Result};
 use circuit_base::{
-    expand_node::{expand_node, MapReplaceExpander, ReplaceMapRc},
+    expand_node::{expand_node, replace_expand_map, MapReplaceExpander, ReplaceMapRc},
     module::{are_args_used, is_intersecting_free_syms, SymbolSetRc},
     CircuitNode, CircuitRc, Module, ModuleArgSpec, ModuleSpec, Rearrange, Symbol,
 };
@@ -17,7 +17,7 @@ use circuit_rewrites::{
     },
 };
 use get_update_node::{AnyFound, IterativeMatcher, IterativeMatcherRc, MatcherData};
-use itertools::{izip, multiunzip};
+use itertools::{izip, multiunzip, Itertools};
 use macro_rules_attribute::apply;
 use pyo3::{
     exceptions::{PyRuntimeError, PyValueError},
@@ -25,7 +25,7 @@ use pyo3::{
     PyObject,
 };
 use rr_util::{
-    cached_method,
+    cached_lambda, cached_method,
     caching::FastUnboundedCache,
     eq_by_big_hash::EqByBigHash,
     fn_struct, impl_eq_by_big_hash,
@@ -38,7 +38,7 @@ use rr_util::{
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use thiserror::Error;
-use uuid::uuid;
+use uuid::{uuid, Uuid};
 
 #[pyfunction]
 #[pyo3(signature=(
@@ -680,13 +680,9 @@ impl ModulePusher {
         if push_or_get {
             assert_eq!(extra_replacements.len(), 0);
         }
-        let (new_get, found_get) = get
-            .match_iterate(circuit.clone())?
-            .unwrap_or_same(get.clone());
+        let (found_get, all_finished, new_get) = get.match_iterate_continue(circuit.clone())?;
 
         let is_get = mode == PushDownMode::Get;
-        let all_finished = new_get.all_finished();
-        let new_get = new_get.clone().per_child_with_term(circuit.num_children());
         // before get_fin for borrow check reasons
         let no_child_will_get = push_or_get
             && !circuit
@@ -744,10 +740,8 @@ impl ModulePusher {
             ));
         }
 
-        let (new_skip_module, found_skip_module) = skip_module
-            .match_iterate(circuit.clone())?
-            .unwrap_or_same(skip_module);
-        let new_skip_module = new_skip_module.per_child_with_term(circuit.num_children());
+        let (found_skip_module, _, new_skip_module) =
+            skip_module.match_iterate_continue(circuit.clone())?;
 
         let mut rec = |self_: &mut Self, c, get, skip, items, extra_replacements: &_| match self_
             .push_down_modules_rec(c, get, skip, mode, items, extra_replacements)?
@@ -811,14 +805,19 @@ impl ModulePusher {
                     .zip(node_get_skip)
                     // don't use rec to avoid adding nodes to out_set
                     .map(|(c, (get, skip))| {
-                        let out = unwrap!(self_.push_down_modules_rec(
-                            c.clone(),
-                            get,
-                            skip,
-                            mode,
-                            &Default::default(), // no items (only difference from recursive call above)
-                            extra_replacements,
-                        )?.0, CircOrGet::Circ);
+                        let out = unwrap!(
+                            self_
+                                .push_down_modules_rec(
+                                    c.clone(),
+                                    get,
+                                    skip,
+                                    mode,
+                                    &Default::default(), // no items (only difference from recursive call above)
+                                    extra_replacements,
+                                )?
+                                .0,
+                            CircOrGet::Circ
+                        );
                         assert_eq!(
                             out.ndim(),
                             c.ndim(),
@@ -848,13 +847,13 @@ impl ModulePusher {
             )?;
             if mode == PushDownMode::PushOverride {
                 let out = Some(
-                    m.map_children_unwrap_idxs(|i| {
-                        if i == 0 {
-                            out.clone().unwrap()
-                        } else {
-                            new_nodes_flat_orig[i - 1].clone().unwrap()
-                        }
-                    })
+                    m.replace_children(
+                        std::iter::once(out)
+                            .chain(new_nodes_flat_orig)
+                            .map(|x| x.unwrap())
+                            .collect(),
+                    )
+                    .unwrap()
                     .rc(),
                 );
                 (out, Default::default(), false)
@@ -1428,17 +1427,12 @@ impl SymbolExtractor {
         symbols: &SymbolSetRc,
         traversal: IterativeMatcherRc,
     ) -> Result<(CircuitRc, BTreeMap<Symbol, (CircuitRc, ModuleArgSpec)>)> {
-        let updated = traversal
-            .match_iterate(circ.clone())?
-            .unwrap_or_same(traversal)
-            .0;
+        let (_, finished, traversal_per_child) = traversal.match_iterate_continue(circ.clone())?;
 
-        if updated.all_finished() {
+        if finished {
             return Ok((circ, Default::default()));
         }
 
-        let traversal_per_child = updated.per_child_with_term(circ.num_children());
-
         let mut out_syms = BTreeMap::default();
         let add_sym_circ = |out_syms: &mut BTreeMap<_, (CircuitRc, ModuleArgSpec)>,
                             arg_spec: ModuleArgSpec,
@@ -1725,3 +1719,195 @@ pub enum ExtractSymbolsError {
         free_outer_symbols: Vec<Symbol>,
     },
 }
+
+#[pyfunction]
+pub fn pull_out_of_modules(
+    circuit: CircuitRc,
+    matcher: IterativeMatcherRc,
+    update_new_binds: MaybeUpdate,
+    matcher_new_binds: Option<IterativeMatcherRc>, /* different matcher for new binds? defaults to matcher */
+    module_construct_callback: Option<PyObject>,
+    circuit_to_arg_spec: Option<PyObject>,
+) -> Result<CircuitRc> {
+    let (a, b) = pull_out_of_modules_get(
+        circuit,
+        matcher,
+        update_new_binds,
+        matcher_new_binds,
+        module_construct_callback,
+        circuit_to_arg_spec,
+    )?;
+    if b.len() != 0 {
+        bail!("pull_out_of_modules: still pulling circuits before returning (call pull_out_of_modules_extract to get results, or use a matcher that deposits them earlier) {:?}", b)
+    }
+    Ok(a)
+}
+
+#[pyfunction]
+pub fn pull_out_of_modules_get(
+    circuit: CircuitRc,
+    matcher: IterativeMatcherRc,
+    update_new_binds: MaybeUpdate,
+    matcher_new_binds: Option<IterativeMatcherRc>, /* different matcher for new binds? defaults to matcher */
+    module_construct_callback: Option<PyObject>,
+    circuit_to_arg_spec: Option<PyObject>,
+) -> Result<(CircuitRc, Vec<(CircuitRc, ModuleArgSpec)>)> {
+    #[apply(cached_lambda)]
+    #[key((circ.info().hash, matcher.clone(), matcher_new_binds.clone()),
+          (HashBytes, IterativeMatcherRc, IterativeMatcherRc))]
+    #[use_try]
+    fn recurse(
+        circ: CircuitRc,
+        matcher: IterativeMatcherRc,
+        matcher_new_binds: IterativeMatcherRc,
+    ) -> Result<(CircuitRc, Vec<(CircuitRc, ModuleArgSpec)>)> {
+        let (found, finished, new_matcher) = matcher.match_iterate_continue(circ.clone())?;
+        let (_, _, new_matcher2) = matcher_new_binds
+            .clone()
+            .match_iterate_continue(circ.clone())?;
+        if found {
+            let spec = if let Some(cb) = &circuit_to_arg_spec {
+                pycall!(cb, (circ.clone(),))
+            } else {
+                ModuleArgSpec::new(
+                    Symbol::new(
+                        circ.shape().clone(),
+                        Uuid::from_slice(&circ.info().hash[..16]).unwrap(),
+                        circ.info().name,
+                    ),
+                    true,
+                    true,
+                    false,
+                )
+            };
+            Ok((spec.symbol.clone().into(), vec![(circ, spec)]))
+        } else if finished {
+            Ok((circ, vec![]))
+        } else if let Some(m) = circ.clone().as_module() {
+            let (spec_circ_2, rs) = recurse(
+                m.spec.circuit.clone(),
+                new_matcher[0].clone(),
+                new_matcher2[0].clone(),
+            )?;
+
+            let vars: IndexSet<Symbol> =
+                m.spec.arg_specs.iter().map(|s| s.symbol.clone()).collect();
+
+            let (rs_with_fvs, rs_no_fvs): (Vec<_>, _) = rs
+                .iter()
+                .cloned()
+                .partition(|(x, _)| is_intersecting_free_syms(x, &vars));
+
+            let m_new = if rs_with_fvs.len() > 0 {
+                let rs_fvs_new = itertools::process_results(
+                    rs_with_fvs.iter().map(|(c, y)| -> Result<_> {
+                        let x = replace_expand_map(
+                            c.clone(),
+                            m.spec
+                                .arg_specs
+                                .iter()
+                                .zip(m.args())
+                                .map(|(s, x)| (s.symbol.crc(), x.clone()))
+                                .collect(),
+                        )?;
+                        Ok((
+                            (
+                                update_new_binds.call(x.clone())?.unwrap_or(x),
+                                y.symbol.info().name,
+                            ),
+                            y,
+                        ))
+                    }),
+                    |it| it.into_group_map(),
+                )?
+                .into_iter()
+                .sorted()
+                .collect::<Vec<_>>();
+
+                let spec_circ_3 = circuit_utils::replace_nodes(
+                    spec_circ_2.clone(),
+                    &rs_fvs_new
+                        .iter()
+                        .flat_map(|(_, v)| {
+                            v[1..]
+                                .iter()
+                                .map(|w| (v[0].symbol.info().hash, w.symbol.crc()))
+                        })
+                        .collect(),
+                );
+
+                Module::try_new(
+                    m.args()
+                        .chain(rs_fvs_new.iter().map(|((x, _), _)| x))
+                        .cloned()
+                        .collect(),
+                    ModuleSpec {
+                        circuit: spec_circ_3,
+                        arg_specs: m
+                            .spec
+                            .arg_specs
+                            .iter()
+                            .chain(rs_fvs_new.iter().map(|(_, b)| b[0]))
+                            .cloned()
+                            .collect(),
+                    },
+                    m.info().name,
+                )?
+                .rc()
+            } else {
+                circ.clone()
+            };
+
+            let (_, _, new_matcher2_again) = matcher_new_binds
+                .clone()
+                .match_iterate_continue(m_new.clone())?;
+
+            let mut out = rs_no_fvs;
+            let circ_out = m_new.map_children_enumerate(|i, c| {
+                Ok(if i == 0 {
+                    if rs_with_fvs.len() == 0 {
+                        spec_circ_2.clone()
+                    } else {
+                        c
+                    }
+                } else {
+                    let i = i - 1;
+                    let (x, y) = recurse(
+                        c,
+                        if i < m.num_args() {
+                            new_matcher[i].clone()
+                        } else {
+                            new_matcher2_again[i].clone()
+                        },
+                        new_matcher2_again[i].clone(),
+                    )?;
+                    out.extend(y);
+                    x
+                })
+            })?;
+
+            let circ_out = if rs_with_fvs.len() > 0 && let Some(cb) = &module_construct_callback {
+                pycall!(cb, (circ_out, circ,), anyhow)?
+            } else {
+                circ_out
+            };
+
+            Ok((circ_out, out))
+        } else {
+            let mut out = Vec::new();
+            let circ_out = circ.map_children_enumerate(|i, c| {
+                let (x, y) = recurse(c, new_matcher[i].clone(), new_matcher2[i].clone())?;
+                out.extend(y);
+                Ok(x)
+            })?;
+
+            Ok((circ_out, out))
+        }
+    }
+
+    Ok(recurse(
+        circuit,
+        matcher.clone(),
+        matcher_new_binds.unwrap_or(matcher),
+    )?)
+}
