@@ -1049,10 +1049,8 @@ where
     let (updated_current_matcher_to_pair, found): (Vec<_>, Vec<_>) = current_matcher_to_pair
         .into_iter()
         .map(|m| {
-            m.match_iterate(circ.clone()).map(|out| {
-                let (new, found) = out.unwrap_or_same(m);
-                (new.per_child_with_term(num_children), found)
-            })
+            let (found, _, new) = m.match_iterate_continue(circ.clone())?;
+            Ok((new, found))
         })
         .collect::<Result<Vec<_>>>()?
         .into_iter()
@@ -1095,7 +1093,9 @@ where
             flatten_impl(c, new_matcher.clone(), updated_matchers_to_pair)
         })
         .collect::<Result<Vec<_>>>()?;
-    let x = x.map_children_unwrap_idxs(|i| rec_on_children[i].0.clone());
+    let x = x
+        .replace_children(rec_on_children.iter().map(|x| x.0.clone()).collect())
+        .unwrap();
 
     Ok(Full::finish_after_rec(&rec_on_children, x))
 }
