@@ -2,11 +2,10 @@ use std::{
     cmp::{max, min},
     iter::zip,
     rc::Rc,
-    sync::{Arc, Mutex},
+    sync::Mutex,
 };
 
-use anyhow::{anyhow, bail, Result};
-use cached::proc_macro::cached;
+use anyhow::{bail, Result};
 use num_bigint::BigUint;
 use once_cell::sync::Lazy;
 /// This file optimizes tensor contraction order, like Python's opt_einsum
@@ -20,10 +19,7 @@ use once_cell::sync::Lazy;
 /// 3: for each subgraph, optimize using DP algorithm based on opt_einsum's DP
 /// 4: contract subgraphs together
 use pyo3::prelude::*;
-use pyo3::{
-    exceptions::PyValueError,
-    types::{PyDict, PyList},
-};
+use pyo3::types::PyDict;
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 
 use crate::{
