@@ -43,6 +43,10 @@ pub struct PyUtils {
     pub random_i64: PyObject,
     pub index_union_rebase: PyObject,
     pub canon_index_tensor: PyObject,
+    pub index_apply_dimwise: PyObject,
+    pub join_indices_unsync: PyObject,
+    pub just_join_indices_sync: PyObject,
+    pub scatter: PyObject,
 }
 
 /// misc python utilities
@@ -99,6 +103,10 @@ pub static PY_UTILS: GILLazyPy<PyUtils> = GILLazyPy::new_py(|py| {
         random_i64: get("random_i64"),
         index_union_rebase: get("index_union_rebase"),
         canon_index_tensor: get("canon_index_tensor"),
+        index_apply_dimwise: get("index_apply_dimwise"),
+        join_indices_unsync: get("join_indices_unsync"),
+        just_join_indices_sync: get("just_join_indices_sync"),
+        scatter: get("scatter"),
     }
 });
 
@@ -369,7 +377,7 @@ generate_extra_py_ops!(add, getitem, mul);
 #[derive(Clone)]
 pub struct Tensor {
     tensor: PyObject,
-    shape: Shape, /* cache shape so doesn't have to be recomputed on reconstruct etc (not uber efficient I think) */
+    pub shape: Shape, /* cache shape so doesn't have to be recomputed on reconstruct etc (not uber efficient I think) */
     hash: Option<HashBytes>,
     device_dtype: TorchDeviceDtype,
 }
