@@ -1,7 +1,15 @@
-use std::{cmp::min, fs, io::Write, iter::zip, os::unix::prelude::OsStrExt, sync::mpsc::channel};
+use std::{
+    cmp::min,
+    fs,
+    io::Write,
+    iter::zip,
+    os::unix::prelude::OsStrExt,
+    sync::{mpsc::channel, Mutex},
+};
 
 use anyhow::{anyhow, bail, Context, Result};
 use base16::encode_lower;
+use once_cell::sync::Lazy;
 use pyo3::prelude::*;
 use rustc_hash::FxHashSet as HashSet;
 use threadpool::ThreadPool;
@@ -12,10 +20,11 @@ use crate::{
     util::HashBytes,
 };
 
-pub fn get_tensor_cache_dir() -> String {
+static TENSOR_CACHE_DIR: Lazy<String> = Lazy::new(|| {
     std::env::var("TENSORS_CACHE_DIR")
         .unwrap_or(std::env::var("HOME").unwrap() + &"/tensors_by_hash_cache")
-}
+});
+static UNSYNCED_LIST: Lazy<String> = Lazy::new(|| TENSOR_CACHE_DIR.clone() + "/.unsynced_list");
 
 pub fn get_rrfs_tensor_db_dir() -> String {
     get_rrfs_dir() + "/tensor_db"
@@ -99,8 +108,8 @@ pub fn get_tensor_dir_tree(dir: &str, hash_prefix: &str) -> Result<Tensor> {
 pub fn get_filename_dir_tree(dir: &str, hash_prefix: &str) -> Result<(String, String)> {
     let (wholes, part) = split_prefix(&hash_prefix);
     let look_dir = dir.to_owned() + "/" + &wholes.join("/");
-    let files =
-        fs::read_dir(&look_dir).context(format!("tensors dir doesnt exist {}", &look_dir))?;
+    let files = fs::read_dir(&look_dir)
+        .with_context(|| format!("tensors dir doesnt exist {}", &look_dir))?;
     let matching_entries: Vec<_> = files
         .filter(|x| &x.as_ref().unwrap().file_name().as_bytes()[..part.len()] == part.as_bytes())
         .collect();
@@ -166,7 +175,7 @@ pub fn register_tensor_unsynced(hash: &str) -> Result<()> {
             .write(true)
             .append(true)
             .create(true)
-            .open(get_tensor_cache_dir() + "/.unsynced_list")
+            .open(&*UNSYNCED_LIST)
             .unwrap(),
         "{}",
         hash
@@ -176,10 +185,9 @@ pub fn register_tensor_unsynced(hash: &str) -> Result<()> {
 
 #[pyfunction(signature=(tensor, force = false))]
 pub fn save_tensor(tensor: Tensor, force: bool) -> Result<bool> {
-    let cache_dir = get_tensor_cache_dir();
-    fs::create_dir_all(&cache_dir)?;
+    fs::create_dir_all(&*TENSOR_CACHE_DIR)?;
 
-    let result = write_tensor_to_dir_tree(&cache_dir, tensor.clone(), force);
+    let result = write_tensor_to_dir_tree(&*TENSOR_CACHE_DIR, tensor.clone(), force);
     if force{
         write_tensor_to_dir_tree(&get_rrfs_tensor_db_dir(), tensor.clone(), force)?;
     }else if let Ok(did_write) = result && did_write{
@@ -196,35 +204,42 @@ fn copy_without_setting_permissions(local_filename: &str, remote_filename: &str)
     std::io::copy(&mut reader, &mut writer).context("writing contents failed")?;
     Ok(())
 }
+
 /// this is only useful bc querying multiple tensors in parallel is much faster than in series
 pub fn ensure_all_tensors_local(hashes: Vec<String>) -> Result<()> {
+    static THREAD_POOL: Lazy<Mutex<ThreadPool>> = Lazy::new(|| {
+        Mutex::new({
+            let n_workers = 30;
+            ThreadPool::new(n_workers)
+        })
+    });
     // println!("loading {} tensors", hashes.len());
-    let n_workers = 30;
     let nitems = hashes.len();
-    let pool = ThreadPool::new(n_workers);
     let (tx, rx) = channel();
-    for hash in hashes {
-        let tx = tx.clone();
-        pool.execute(move || {
-            let result = ensure_prefix_local(&hash).map(|_| ());
-            // println!("did");
-            tx.send(result).unwrap()
-        });
+    {
+        let pool = THREAD_POOL.lock().unwrap();
+        for hash in hashes {
+            let tx = tx.clone();
+            pool.execute(move || {
+                let result = ensure_prefix_local(&hash).map(|_| ());
+                // println!("did");
+                tx.send(result).unwrap()
+            });
+        }
     }
     rx.iter().take(nitems).collect::<Result<()>>()
 }
 
 #[pyfunction(signature=(parallelism = 15))]
 pub fn sync_all_unsynced_tensors(parallelism: usize) -> Result<()> {
-    let unsynced_list_path = get_tensor_cache_dir() + "/.unsynced_list";
-    if !fs::try_exists(&unsynced_list_path)? {
+    if !fs::try_exists(&*UNSYNCED_LIST)? {
         println!(
             "didnt find any unsynced list (~/tensors_by_hash_cache/.unsynced_list), returning"
         );
         return Ok(());
     }
     println!("starting to read list");
-    let all_unsynced = fs::read_to_string(&unsynced_list_path)?;
+    let all_unsynced = fs::read_to_string(&*UNSYNCED_LIST)?;
     println!("read list {}", all_unsynced.len());
     let lines: Vec<String> = all_unsynced
         .lines()
@@ -247,10 +262,9 @@ pub fn sync_specific_tensors(tensor_hash_prefixes: Vec<String>, parallelism: usi
         let tx = tx.clone();
         pool.execute(move || {
             tx.send((move || {
-                let local_filename =
-                    get_filename_dir_tree(&get_tensor_cache_dir(), &hash_prefix_str)?.0;
+                let local_filename = get_filename_dir_tree(&*TENSOR_CACHE_DIR, &hash_prefix_str)?.0;
                 let remote_filename =
-                    local_filename.replace(&get_tensor_cache_dir(), &get_rrfs_tensor_db_dir());
+                    local_filename.replace(&*TENSOR_CACHE_DIR, &get_rrfs_tensor_db_dir());
                 let remote_filename_dir: Vec<_> = remote_filename.split("/").collect();
                 let remote_filename_dir = remote_filename_dir
                     [..remote_filename_dir.len().checked_add_signed(-1).unwrap()]
@@ -268,15 +282,14 @@ pub fn sync_specific_tensors(tensor_hash_prefixes: Vec<String>, parallelism: usi
         .take(tensor_hash_prefixes.len())
         .collect::<Result<Vec<()>>>()?;
 
-    let unsynced_list_path = get_tensor_cache_dir() + "/.unsynced_list";
-    if !fs::try_exists(&unsynced_list_path)? {
+    if !fs::try_exists(&*UNSYNCED_LIST)? {
         println!(
             "didnt find any unsynced list (~/tensors_by_hash_cache/.unsynced_list), returning"
         );
         return Ok(());
     }
 
-    let all_unsynced = fs::read_to_string(&unsynced_list_path)?;
+    let all_unsynced = fs::read_to_string(&*UNSYNCED_LIST)?;
     println!("read list {}", all_unsynced.len());
     let still_unsynced_lines: Vec<String> = all_unsynced
         .lines()
@@ -286,16 +299,15 @@ pub fn sync_specific_tensors(tensor_hash_prefixes: Vec<String>, parallelism: usi
         .map(|x| x.to_owned() + "\n")
         .collect();
 
-    fs::write(&unsynced_list_path, still_unsynced_lines.join("")).context("write failed")
+    fs::write(&*UNSYNCED_LIST, still_unsynced_lines.join("")).context("write failed")
 }
 
 #[pyfunction]
 pub fn get_tensor_prefix(prefix: &str) -> Result<Tensor> {
-    let cache_dir = get_tensor_cache_dir();
-    get_tensor_dir_tree(&cache_dir, &prefix).or_else(|_e| {
+    get_tensor_dir_tree(&*TENSOR_CACHE_DIR, &prefix).or_else(|_e| {
         let result = get_tensor_dir_tree(&get_rrfs_tensor_db_dir(), &prefix);
         if let Ok(t) = &result {
-            write_tensor_to_dir_tree(&cache_dir, t.clone(), false)?;
+            write_tensor_to_dir_tree(&*TENSOR_CACHE_DIR, t.clone(), false)?;
         }
         result
     })
@@ -303,13 +315,12 @@ pub fn get_tensor_prefix(prefix: &str) -> Result<Tensor> {
 
 #[pyfunction]
 pub fn ensure_prefix_local(prefix: &str) -> Result<()> {
-    let cache_dir = get_tensor_cache_dir();
-    get_filename_dir_tree(&cache_dir, &prefix)
+    get_filename_dir_tree(&*TENSOR_CACHE_DIR, &prefix)
         .map(|_| ())
         .or_else(|_e| {
             let remote_filename = get_filename_dir_tree(&get_rrfs_tensor_db_dir(), prefix)?.0;
             let local_filename =
-                remote_filename.replace(&get_rrfs_tensor_db_dir(), &get_tensor_cache_dir());
+                remote_filename.replace(&get_rrfs_tensor_db_dir(), &*TENSOR_CACHE_DIR);
             let local_filename_dir: Vec<_> = local_filename.split("/").collect();
             let local_filename_dir = local_filename_dir
                 [..local_filename_dir.len().checked_add_signed(-1).unwrap()]
@@ -328,8 +339,7 @@ pub fn migrate_tensors_from_old_dir(dir: &str) -> Result<()> {
         .map(|x| x.unwrap().file_name().to_str().unwrap().to_owned())
         .collect();
     print!("have list");
-    let list_dir = get_tensor_cache_dir() + "/.unsynced_list";
-    let all_unsynced = fs::read_to_string(&list_dir)?;
+    let all_unsynced = fs::read_to_string(&*UNSYNCED_LIST)?;
     let lines: HashSet<_> = all_unsynced[..all_unsynced.len() - 1].lines().collect();
 
     for file in files {
