@@ -36,6 +36,94 @@ use crate::{
 };
 pub type Shape = Sv<[usize; 4]>; // TODO: maybe this should be struct which preserves invariant that dim shapes >= 0?
 
+/// casts tensors whith None device/dtype to specified device/dtype
+pub fn cast_tensors(tensors: &[Tensor], out_device_dtype: TorchDeviceDtype) -> Vec<Tensor> {
+    tensors
+        .iter()
+        .map(|t| {
+            if t.device_dtype() != out_device_dtype {
+                pycall!(PY_UTILS.cast_tensor, (t.clone(), out_device_dtype))
+            } else {
+                t.clone()
+            }
+        })
+        .collect()
+}
+
+pub fn upcast_tensor_device_dtypes(tensors: &[Tensor]) -> Vec<Tensor> {
+    let device = upcast_devices_if_needed(
+        &tensors
+            .iter()
+            .map(|z| z.device_dtype().device)
+            .collect::<Vec<_>>(),
+    );
+    let dtype = upcast_dtypes_if_needed(
+        &tensors
+            .iter()
+            .map(|z| z.device_dtype().dtype)
+            .collect::<Vec<_>>(),
+    );
+    if device.is_some() || dtype.is_some() {
+        return cast_tensors(
+            tensors,
+            TorchDeviceDtypeOp { device, dtype }.unwrap_or_defaults(),
+        );
+    }
+    tensors.to_owned()
+}
+
+fn upcast_devices_if_needed(devices: &[TorchDevice]) -> Option<TorchDevice> {
+    if devices.len() == 0 || devices[1..].iter().all(|z| z == &devices[0]) {
+        return None;
+    }
+    devices
+        .iter()
+        .cloned()
+        .filter(|z| *z != TorchDevice::Cpu)
+        .next()
+}
+fn upcast_dtypes_if_needed(dtypes: &[TorchDtype]) -> Option<TorchDtype> {
+    if dtypes.len() == 0 || dtypes[1..].iter().all(|z| z == &dtypes[0]) {
+        return None;
+    }
+    dtypes
+        .iter()
+        .cloned()
+        .filter(|z| *z != TorchDtype::float64)
+        .next()
+}
+
+#[pyfunction]
+#[pyo3(name = "upcast_tensor_devices")]
+pub fn upcast_tensor_devices_py(tensors: Vec<Tensor>) -> Vec<Tensor> {
+    upcast_tensor_devices(&tensors)
+}
+#[pyfunction]
+#[pyo3(name = "upcast_tensor_device_dtypes")]
+pub fn upcast_tensor_device_dtypes_py(tensors: Vec<Tensor>) -> Vec<Tensor> {
+    upcast_tensor_device_dtypes(&tensors)
+}
+
+pub fn upcast_tensor_devices(tensors: &[Tensor]) -> Vec<Tensor> {
+    if let Some(device) = upcast_devices_if_needed(
+        &tensors
+            .iter()
+            .cloned()
+            .map(|z| z.device_dtype().device)
+            .collect::<Vec<_>>(),
+    ) {
+        return cast_tensors(
+            tensors,
+            TorchDeviceDtypeOp {
+                dtype: None,
+                device: Some(device),
+            }
+            .unwrap_or_defaults(),
+        );
+    }
+    tensors.to_owned()
+}
+
 pub fn eq_or_symbolic(x: &[usize], y: &[usize]) -> bool {
     x.len() == y.len()
         && x.iter().zip(y).all(|(a, b)| {
@@ -235,6 +323,10 @@ pub fn check_canon_idxs(count: usize, ints: &[i64]) -> Result<Vec<usize>> {
         .collect()
 }
 
+pub fn check_canon_axes(rank: usize, axes: &[i64]) -> Result<Vec<usize>> {
+    check_canon_idxs(rank, axes).context("axis out of bounds")
+}
+
 // we could bit pack if we wanted...
 #[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
 pub struct Slice {
@@ -297,7 +389,7 @@ impl Slice {
         }
     }
 
-    fn is_identity(self, l: usize) -> bool {
+    pub fn is_identity(self, l: usize) -> bool {
         self.start_u(l) == 0 && self.stop_u(l) == l
     }
 }
@@ -531,21 +623,32 @@ impl TensorAxisIndex {
     }
 
     // assumes self canonicalized
-    pub fn shrink_base_uslice(&self, uslice: &USlice) -> Self {
+    fn offset(&self, b: i64) -> Self {
         match self {
-            TensorAxisIndex::Single(single) => {
-                TensorAxisIndex::Single(single - uslice.start as i64)
+            TensorAxisIndex::Slice(i) => {
+                let i2: USlice = Into::<Option<USlice>>::into(*i).unwrap();
+                TensorAxisIndex::Slice(
+                    USlice {
+                        start: (i2.start as i64 + b).try_into().unwrap(),
+                        stop: (i2.stop as i64 + b).try_into().unwrap(),
+                    }
+                    .into(),
+                )
             }
-            TensorAxisIndex::Slice(slice) => TensorAxisIndex::Slice(Slice {
-                start: Some(slice.start.unwrap() - uslice.start as i64),
-                stop: Some(slice.stop.unwrap() - uslice.start as i64),
+            TensorAxisIndex::Single(i) => {
+                TensorAxisIndex::Single(TryInto::<usize>::try_into(i + b).unwrap() as i64)
+            }
+            TensorAxisIndex::Tensor(t) => Python::with_gil(|py| {
+                TensorAxisIndex::Tensor(t.0.clone().py_add(py, b).unwrap().try_into().unwrap())
             }),
-            TensorAxisIndex::Tensor(_tensor) => {
-                unimplemented!();
-            }
         }
     }
 
+    // assumes self canonicalized
+    pub fn shrink_base_uslice(&self, uslice: &USlice) -> Self {
+        self.offset(-(uslice.start as i64))
+    }
+
     pub fn canonicalize(&self, l: usize) -> Self {
         match self {
             TensorAxisIndex::Single(i) => TensorAxisIndex::Single((*i + l as i64) % l as i64),
@@ -603,6 +706,14 @@ impl TensorAxisIndex {
                 .unwrap()
         }))
     }
+
+    pub fn axis_len(&self, l: usize) -> usize {
+        match self {
+            TensorAxisIndex::Single(_) => 0,
+            TensorAxisIndex::Slice(s) => s.to_uslice(l).length(),
+            TensorAxisIndex::Tensor(t) => t.0.shape[0],
+        }
+    }
 }
 
 impl IntoPy<PyObject> for TensorAxisIndex {
@@ -692,7 +803,29 @@ impl TensorIndex {
             .collect()
     }
 
-    pub fn is_identity(&self, shape: &Shape) -> bool {
+    pub fn edges_to_none(&self, shape: &Shape) -> Self {
+        TensorIndex(
+            zip(&self.0, shape)
+                .map(|(ax, l)| match ax {
+                    TensorAxisIndex::Slice(slice) => TensorAxisIndex::Slice(Slice {
+                        start: if slice.start_u(*l) == 0 {
+                            None
+                        } else {
+                            slice.start
+                        },
+                        stop: if slice.stop_u(*l) == *l {
+                            None
+                        } else {
+                            slice.stop
+                        },
+                    }),
+                    _ => ax.clone(),
+                })
+                .collect(),
+        )
+    }
+
+    pub fn is_identity(&self, shape: &[usize]) -> bool {
         zip(&self.0, shape).all(|(idx, &l)| idx.is_identity(l))
     }
 
