@@ -285,7 +285,7 @@ raw_attention_str = f"""
   'qk_with_rot.re' [{SEQ}, 2*2*{HALF_ROTARY_DIM}] Rearrange q s:2 e:2 rh -> q (s:2 e:2 rh)
     'qk_with_rot.cat' [{SEQ}, 2, 2, {HALF_ROTARY_DIM}] Concat 1
       'qk_with_rot.q_rot' Rearrange q e rh -> q 1 e rh
-        'qk_rot' [{SEQ}, 2, {HALF_ROTARY_DIM}] GeneralFunction apply_rotary_pos_emb at interp.circuit.interop_rust.generalfuncs.rotary_pos_emb:ApplyRotaryPosEmb
+        'qk_rot' [{SEQ}, 2, {HALF_ROTARY_DIM}] GeneralFunction apply_rotary_pos_emb at rust_circuit.generalfuncs.rotary_pos_emb:ApplyRotaryPosEmb
           'qk_rot.inp' Index [:, 0]
             'qk_split' [{SEQ}, 2, 2, {HALF_ROTARY_DIM}] Rearrange q (s:2 e:2 rh) -> q s:2 e:2 rh
               'qk_split.set' [{SEQ}, 2*2*{HALF_ROTARY_DIM}] SetSymbolicShape
@@ -1005,16 +1005,31 @@ def load_transformer_model_string(
 
     circs = info.params.get().circs
     added = add_new_circs(s, circs, parser=parser)
-    from interp.model.model_loading import MODEL_CLASS_STR_TO_MODEL_AND_TOKENIZER_FNS
 
-    tokenizer_fn = MODEL_CLASS_STR_TO_MODEL_AND_TOKENIZER_FNS[info.model_class][1]
+    # copied out of model_loading to work in open source version
+    try:
+        from interp.model.model_loading import MODEL_CLASS_STR_TO_MODEL_AND_TOKENIZER_FNS
 
-    return (added, tokenizer_fn(), info)
+        tokenizer = MODEL_CLASS_STR_TO_MODEL_AND_TOKENIZER_FNS[info.model_class][1]()
+    except ImportError:
+        if info.model_class in {"GPT", "GPTBeginEndToks"}:
+            import transformers
+
+            tokenizer = transformers.GPT2TokenizerFast.from_pretrained("gpt2")
+            tokenizer._add_tokens(["[END]"])
+            tokenizer.pad_token = "[END]"
+            if info.model_class == "GPTBeginEndToks":
+                tokenizer._add_tokens(["[BEGIN]"])
+                tokenizer.eos_token = "[END]"
+        else:
+            raise
+
+    return (added, tokenizer, info)
 
 
 def get_model_path(model_id: str):
     # we should plausibly eventually remove current circ_models/ and replace with circ_models2/
-    from interp.tools.rrfs import RRFS_DIR
+    from rust_circuit.rrfs import RRFS_DIR
 
     return f"{RRFS_DIR}/circ_models2/{model_id}.circ"
 
@@ -1028,12 +1043,11 @@ def load_model_id(model_id: str, parser: Optional[Parser] = None) -> tuple[dict[
 token_upcast_module = make_spec(
     P(
         """
-'upcast_toks' GeneralFunction cast_from_{device:None,dtype:None}_to_{device:None,dtype:int64}
-  'upcast_toks.sub' Add
+'upcast_toks.sub' Add
+  'upcast_toks' GeneralFunction cast_from_{device:None,dtype:None}_to_{device:None,dtype:int64}
     'upcast_toks.int16_toks' [] Symbol
-    'upcast_toks.signed_int16_min' [] Scalar 32768 # int16 min, torch doesn't support uint16 :(
+  'upcast_toks.signed_int16_min' [] Scalar 32768 # int16 min, torch doesn't support uint16 :(
 """
     ),
     [P("'upcast_toks.int16_toks' [] Symbol")],
 )
-print(token_upcast_module)
