@@ -2,7 +2,6 @@ from datetime import timedelta
 from functools import partial
 from typing import Any, Callable, List, Sequence, Set, Tuple, cast
 
-import attrs
 import hypothesis
 import hypothesis.extra.numpy as st_np
 import hypothesis.strategies as st
@@ -10,22 +9,22 @@ import torch
 from hypothesis import example, given
 
 import interp.circuit.testing.strategies as st_c
-from interp.circuit.circuit import Circuit, CircuitTypeError, MemoizedFn
-from interp.circuit.circuit_compiler.util import FrozenDict, TorchAxisIndex
-from interp.circuit.circuit_utils import evaluate_fn
-from interp.circuit.computational_node import Add, Concat, Einsum, Index, UnaryRearrange, dedup_with_order
-from interp.circuit.constant import ArrayConstant, FloatConstant, One, Zero
-from interp.circuit.sample_transform import RunDiscreteVarAllSpec, SampleTransform
+import rust_circuit as rc
+from interp.circuit.circuit_compiler.util import TorchAxisIndex
 from interp.circuit.testing.topdown_circuit import CircuitProbs as CP
-from interp.circuit.testing.topdown_circuit import st_Circuit
 from interp.circuit.testing.utils import mark_not_interesting_if, target_diff
-from interp.circuit.var import AutoTag
+from rust_circuit import Add
 from rust_circuit import Add as rAdd
+from rust_circuit import Array
+from rust_circuit import Circuit
 from rust_circuit import Circuit as rCircuit
+from rust_circuit import Concat
+from rust_circuit import Einsum
 from rust_circuit import Einsum as rEinsum
-from rust_circuit import Matcher, Parser
+from rust_circuit import Index, Matcher, Parser, Rearrange
+from rust_circuit import Scalar
 from rust_circuit import Scalar as rScalar
-from rust_circuit import apply_in_traversal, new_traversal, replace_outside_traversal_symbols, simp
+from rust_circuit import Tag, apply_in_traversal, new_traversal, replace_outside_traversal_symbols, simp
 from rust_circuit.algebric_rewrite import (
     ImpossibleRewriteError,
     RewriteError,
@@ -66,15 +65,14 @@ from rust_circuit.algebric_rewrite import (
     split_to_concat_for_batch,
     zero_empty_add,
 )
-from rust_circuit.interop_rust import py_to_rust, rust_to_py
 
-from .test_rust_rewrite import deterministic_rearrange_noneing
+from .util import dedup_with_order, st_Circuit
 
 P = Parser()
 
 
 def is_zero(c: Circuit):
-    return isinstance(c, Zero) or (isinstance(c, FloatConstant) and c.value == 0)
+    return c.is_scalar() and c.cast_scalar().is_zero()
 
 
 @st.composite
@@ -113,7 +111,7 @@ def st_index_groups(draw: st.DrawFn, length: int) -> List[List[int]]:
 @st.composite
 def st_index_groups_and_circuit(
     draw: st.DrawFn, must_be_explicitly_computable: bool = True
-) -> Tuple[List[List[int]], int, Circuit]:
+) -> Tuple[List[List[int]], int, rCircuit]:
     st_shapes = st_np.array_shapes(min_dims=1, max_dims=4, max_side=8)
     st_circuit = st_Circuit(
         st_shapes,
@@ -128,12 +126,12 @@ def st_index_groups_and_circuit(
     circuit = draw(st_circuit)
     assert isinstance(circuit, Einsum) and len(circuit.args) > 0
 
-    concat_idxs = [i for i, c in enumerate(circuit.all_input_circuits()) if isinstance(c, Concat)]
+    concat_idxs = [i for i, c in enumerate(circuit.children) if isinstance(c, Concat)]
     hypothesis.assume(len(concat_idxs) > 0, "no concat indices")
     which = concat_idxs[draw(st.integers(min_value=0, max_value=len(concat_idxs) - 1))]
 
-    index_groups_target: Concat = cast(Concat, (circuit.args[which][0]))
-    index_groups = draw(st_index_groups(len(index_groups_target.circuits)))
+    index_groups_target: Concat = circuit.children[which].cast_concat()
+    index_groups = draw(st_index_groups(index_groups_target.num_children))
     return index_groups, which, circuit
 
 
@@ -151,7 +149,9 @@ def test_auto_split_einsum_concat(groups_and_circuit):
 
 @mark_not_interesting_if(RewriteError)
 # use Any for `fn` type to avoid errors
-def check_transform_equality(fn: Callable[[Any], rCircuit], circuit: Circuit, sample: bool = False) -> Circuit:
+def check_transform_equality(
+    fn: Callable[[Any], rCircuit], circuit: Circuit, sample: bool = False, f64: bool = True
+) -> Circuit:
     """
     Checks that when we apply the rewrite `fn` to `circuit` the evaluated circuit has the same value
 
@@ -162,24 +162,25 @@ def check_transform_equality(fn: Callable[[Any], rCircuit], circuit: Circuit, sa
       :class:`NotImplementedError`. This is true for all rewrites *except* :func:`rearrange_muls`. In that case, many successful rearranges
       would leave the tree of Einsums equal, and so this doesn't make sense.
     """
-    circuit_rust = deterministic_rearrange_noneing(py_to_rust(circuit))
-    trans_rust = fn(circuit_rust)
-    trans = rust_to_py(trans_rust)
+    if f64:
+        circuit = rc.cast_circuit(circuit, rc.TorchDeviceDtypeOp(dtype="float64"))
+    trans = fn(circuit)
 
     assert trans.shape == circuit.shape
-    assert trans_rust != circuit_rust, "The rewrite should raise a RewriteError instead of nooping."
-    evaluator = MemoizedFn(evaluate_fn(torch.float64))
+    assert trans != circuit, "The rewrite should raise a RewriteError instead of nooping."
 
     if sample:
-        transform = SampleTransform(RunDiscreteVarAllSpec.create_full_from_circuits(trans, circuit))
-        trans_eval = transform(trans)
-        circuit_eval = transform(circuit)
+        sampler = rc.Sampler(rc.RunDiscreteVarAllSpec.create_full_from_circuits(trans, circuit))
+        trans_eval = sampler(trans)
+        circuit_eval = sampler(circuit)
     else:
         trans_eval = trans
         circuit_eval = circuit
 
-    actual, expected = (evaluator(trans_eval), evaluator(circuit_eval))
-    target_diff(actual, expected, evaluator=evaluator)
+    hypothesis.note(circuit_eval.repr())
+    hypothesis.note(trans_eval.repr())
+    actual, expected = (trans_eval.evaluate(), circuit_eval.evaluate())
+    target_diff(actual.type(expected.dtype), expected)  # , evaluator=evaluator)
     return trans
 
 
@@ -190,23 +191,18 @@ def check_transform_equality(fn: Callable[[Any], rCircuit], circuit: Circuit, sa
         must_be_explicitly_computable=True,
         max_growth_steps=10,
         # Sample more relevant nodes at depth 0 and more zeros at depth 1
-        probs_default=CP.kw(all=1, Zero=5),
-        probs_per_depth=[CP.kw(all=0, Add=10, Einsum=10, UnaryRearrange=10, Index=10)],
+        probs_default=CP.kw(all=1, Scalar=(5, dict(values=[0.0], also_any=True))),
+        probs_per_depth=[CP.kw(all=0, Add=10, Einsum=10, Rearrange=10, Index=10)],
     )
 )
 def test_eliminate_zeros(circuit):
     """
-    Check that if the top level is a :class:`Einsum` or :class:`UnaryRearrange` or :class:`Index` or :class:`Add`, it has been
+    Check that if the top level is a :class:`Einsum` or :class:`Rearrange` or :class:`Index` or :class:`Add`, it has been
     replaced by a zero (and the shape is still the same). Also exact float equality.
     """
     transformed_circuit = check_transform_equality(eliminate_zeros, circuit)
 
-    children = (
-        list(transformed_circuit.items.keys())
-        if isinstance(transformed_circuit, Add)
-        else list(transformed_circuit.children())
-    )
-    assert not any(is_zero(c) for c in children)
+    assert not any(is_zero(c) for c in transformed_circuit.children)
 
 
 @given(
@@ -215,7 +211,7 @@ def test_eliminate_zeros(circuit):
         must_be_explicitly_computable=True,
         max_growth_steps=10,
         probs_default=CP.kw(all=1),
-        probs_per_depth=[CP.kw(all=0, Einsum=1), CP.kw(all=1, One=20)],
+        probs_per_depth=[CP.kw(all=0, Einsum=1), CP.kw(all=1, Scalar=(20, dict(values=[1.0], also_any=True)))],
     )
 )
 @hypothesis.settings(deadline=timedelta(milliseconds=400))
@@ -228,7 +224,7 @@ def test_drop_mul_ones(circuit):
         st_np.array_shapes(min_dims=0),
         must_be_explicitly_computable=True,
         max_growth_steps=3,
-        probs_per_depth=[CP.kw(all=0, UnaryRearrange=1), CP.kw(all=0, One=1, Zero=1, FloatConstant=0)],
+        probs_per_depth=[CP.kw(all=0, Rearrange=1), CP.kw(all=0, Scalar=(1, dict(values=[0.0, 1.0])))],
     )
 )
 def test_rearrange_of_const_to_const(circuit):
@@ -250,19 +246,18 @@ def test_flatten_adds(circuit):
     Check that if the top-most circuit is an Add, and we apply `flatten_adds`
     many times, we can get rid of the Add children.
     """
-    evaluator = MemoizedFn(evaluate_fn(torch.float64))
-    evaluator(circuit)  # Count the number of nodes in `circuit` by caching them
-    n_nodes = len(evaluator.cache)  # An upper bound on the number of `flatten_adds` transforms we need
+    circuit.evaluate()
+    n_nodes = len(rc.all_children(circuit))  # An upper bound on the number of `flatten_adds` transforms we need
 
-    trans = py_to_rust(circuit)
+    trans = circuit
     n_iter = 0
     while isinstance(trans, rAdd) and any(isinstance(c, rAdd) for c in trans.children):
         try:
             trans = flatten_adds(trans)
         except ImpossibleRewriteError as e:
             raise RuntimeError(f"There are child adds, should not have raised {e}")
-        assert rust_to_py(trans).shape == circuit.shape
-        torch.testing.assert_close(evaluator(rust_to_py(trans)), evaluator(circuit))
+        assert trans.shape == circuit.shape
+        torch.testing.assert_close(trans.evaluate(), circuit.evaluate())
         n_iter += 1
         assert n_iter <= n_nodes, "flatten_adds is not getting rid of the `Add`s"
 
@@ -276,10 +271,11 @@ def test_flatten_adds(circuit):
         must_be_explicitly_computable=True,
         max_growth_steps=3,
         probs_default=CP.kw(all=1),
-        probs_per_depth=[CP.kw(all=0, UnaryRearrange=(1, dict(do_squeeze=True)), Einsum=0)],
+        probs_per_depth=[CP.kw(all=0, Rearrange=(1, dict(do_squeeze=True)), Einsum=0)],
+        rust=True,
     )
 )
-def test_explicit_squeeze(circuit: UnaryRearrange):
+def test_explicit_squeeze(circuit: rc.Rearrange):
     check_transform_equality(explicit_squeeze, circuit)
 
 
@@ -294,7 +290,7 @@ def test_explicit_squeeze(circuit: UnaryRearrange):
         must_be_explicitly_computable=True,
         max_growth_steps=3,
         probs_default=CP.kw(all=1),
-        probs_per_depth=[CP.kw(all=0, UnaryRearrange=1, Einsum=1)],
+        probs_per_depth=[CP.kw(all=0, Rearrange=1, Einsum=1)],
     )
 )
 def test_remove_noop_rearrange(circuit):
@@ -313,7 +309,7 @@ def test_remove_noop_rearrange(circuit):
 )
 def test_remove_single_concat(circuit: Circuit, keep_name: bool, axis: int):
     assert circuit.ndim > 0
-    add_circuit = Concat((circuit,), axis=axis % circuit.ndim)
+    add_circuit = Concat(circuit, axis=axis % circuit.ndim)
     trans = check_transform_equality(partial(remove_single_concat, keep_name=keep_name), add_circuit)
     if keep_name:
         assert trans.name == add_circuit.name
@@ -321,26 +317,27 @@ def test_remove_single_concat(circuit: Circuit, keep_name: bool, axis: int):
 
 @st.composite
 def generate_einsum_with_squeeze(draw: st.DrawFn) -> Tuple[int, int, Einsum]:
-    ein = Einsum.unwrap(
-        draw(
-            st_Circuit(
-                st_np.array_shapes(min_dims=0, max_dims=7),
-                must_be_explicitly_computable=True,
-                max_growth_steps=3,
-                probs_default=CP.kw(all=1, UnaryRearrange=20),
-                probs_per_depth=[CP.kw(all=0, Einsum=(1, dict(min_n_children=1)))],
-            )
+    ein = draw(
+        st_Circuit(
+            st_np.array_shapes(min_dims=0, max_dims=7),
+            must_be_explicitly_computable=True,
+            max_growth_steps=3,
+            probs_default=CP.kw(all=1, Rearrange=20),
+            probs_per_depth=[CP.kw(all=0, Einsum=(1, dict(min_n_children=1)))],
         )
-    )
+    ).cast_einsum()
     to_unsqueeze_arg = draw(st.integers(min_value=0, max_value=len(ein.args) - 1))
-    (node, axes), other_args = ein.get_info_by(to_unsqueeze_arg)
+    # (node, axes), other_args = ein.get_info_by(to_unsqueeze_arg)
+    args = ein.args
+    node, axes = args[to_unsqueeze_arg]
     to_unsqueeze_axis = draw(st.integers(min_value=0, max_value=len(axes)))
 
-    new_node = UnaryRearrange.unsqueeze(node, [to_unsqueeze_axis])
-    new_axes = axes[:to_unsqueeze_axis] + (ein.end_num(),) + axes[to_unsqueeze_axis:]
+    new_node = node.unsqueeze([to_unsqueeze_axis])
+    new_axes = axes[:to_unsqueeze_axis] + (ein.next_axis(),) + axes[to_unsqueeze_axis:]
     assert new_node.shape[to_unsqueeze_axis] == 1
+    args[to_unsqueeze_arg] = (new_node, new_axes)
 
-    new_ein = ein.reconstruct_by(to_unsqueeze_arg, new_node, new_axes, other_args)
+    new_ein = Einsum(*args, out_axes=ein.out_axes, name=ein.name)
     return to_unsqueeze_arg, to_unsqueeze_axis, new_ein
 
 
@@ -379,18 +376,16 @@ def st_add_and_subset(draw: st.DrawFn) -> Tuple[Add, Set[Circuit]]:
     see test below)
     """
 
-    circuit: Add = Add.unwrap(
-        draw(
-            st_Circuit(
-                st_np.array_shapes(min_dims=0),
-                must_be_explicitly_computable=True,
-                max_growth_steps=1,
-                probs_per_depth=[CP.kw(all=0, Add=1)],
-                add_must_have_weight_one=True,
-            )
+    circuit: Add = draw(
+        st_Circuit(
+            st_np.array_shapes(min_dims=0),
+            must_be_explicitly_computable=True,
+            max_growth_steps=1,
+            probs_per_depth=[CP.kw(all=0, Add=1)],
+            add_must_have_weight_one=True,
         )
-    )
-    children = list(circuit.items)
+    ).cast_add()
+    children = circuit.children
     assert len(children) == len(set(children))
 
     if len(children) == 0:
@@ -406,44 +401,40 @@ def test_drop_from_add(circuit_and_subset: Tuple[Add, Set[Circuit]]):
     subset: Set[Circuit]
     circuit, subset = circuit_and_subset
     try:
-        rust_children = py_to_rust(circuit).cast_add().children
-        hypothesis.assume(len(set(rust_children)) == len(rust_children))
-        new_circuit = rust_to_py(
-            drop_from_add(
-                py_to_rust(circuit).cast_add(), {cast(rCircuit, py_to_rust(c)) for c in subset}, noop_error=True
-            )
-        )
-    except RewriteError as e:
-        hypothesis.assume(False, repr(e))
-        raise e  # unreachable but makes mypy happy
+        children = circuit.children
+        hypothesis.assume(len(set(children)) == len(children))
+        new_circuit = drop_from_add(circuit, subset, noop_error=True)
+    except RewriteError:
+        hypothesis.reject()
 
     assert len(subset) > 0
 
     assert new_circuit.shape == circuit.shape
     assert new_circuit != circuit, "The rewrite should raise ImpossibleRewriteError instead of nooping"
 
-    if isinstance(new_circuit, UnaryRearrange):
-        new_children = list(new_circuit.node.children())
+    if isinstance(new_circuit, Rearrange):
+        new_children = new_circuit.node.children
     elif is_zero(new_circuit):
         new_children = []
     elif isinstance(new_circuit, Add):
-        new_children = list(new_circuit.children())
+        new_children = new_circuit.children
     else:
         assert False
 
-    assert len(new_children) == len(list(circuit.children())) - len(
+    assert len(new_children) == circuit.num_children - len(
         subset
-    ), f"{len(new_children)} != {len(list(circuit.children()))} - {len(subset)}; {new_circuit}"
+    ), f"{len(new_children)} != {circuit.num_children} - {len(subset)}; {new_circuit}"
 
     for deleted_circuit in subset:
         assert deleted_circuit not in new_children
 
-    old_circuit_with_zeros = attrs.evolve(
-        circuit, items=FrozenDict({k: (0.0 if k in subset else v) for k, v in circuit.items.items()})
+    old_circuit_with_zeros = Add.from_weighted_nodes(
+        *((k, (0.0 if k in subset else v)) for k, v in circuit.to_counts().items())
     )
+    old_circuit_with_zeros = rc.cast_circuit(old_circuit_with_zeros, rc.TorchDeviceDtypeOp(dtype="float64"))
 
-    evaluator = MemoizedFn(evaluate_fn(torch.float64))
-    assert torch.equal(evaluator(old_circuit_with_zeros), evaluator(new_circuit))
+    new_out = new_circuit.evaluate()
+    torch.testing.assert_close(old_circuit_with_zeros.evaluate().type(new_out.dtype), new_out)
 
 
 # this only activates if add weight is 0, which is rare, so don't demand many examples
@@ -462,7 +453,7 @@ def test_drop_zero_add(circuit):
 
 
 @hypothesis.example(
-    circuit=Add(items={}, name="", shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True),
+    circuit=Add(),
 )
 @given(
     circuit=st_Circuit(
@@ -479,18 +470,16 @@ def test_zero_empty_add(circuit: Add):
 
 @st.composite
 def st_distributed_einsum(draw: st.DrawFn):
-    c: Einsum = Einsum.unwrap(
-        draw(
-            st_Circuit(
-                st_np.array_shapes(min_dims=0),
-                must_be_explicitly_computable=True,
-                max_growth_steps=5,
-                probs_default=CP.kw(all=1),
-                probs_per_depth=[CP.kw(all=0, Einsum=(1, dict(min_n_children=2, max_n_children=2)))],
-            )
+    c: Einsum = draw(
+        st_Circuit(
+            st_np.array_shapes(min_dims=0),
+            must_be_explicitly_computable=True,
+            max_growth_steps=5,
+            probs_default=CP.kw(all=1),
+            probs_per_depth=[CP.kw(all=0, Einsum=(1, dict(min_n_children=2, max_n_children=2)))],
         )
-    )
-    assert len(c.all_input_circuits()) == 2
+    ).cast_einsum()
+    assert c.num_children == 2
 
     ((a, a_axes), (b, b_axes)) = c.args
     out_axes = c.out_axes
@@ -508,11 +497,11 @@ def st_distributed_einsum(draw: st.DrawFn):
             )
         )
 
-        c_p = attrs.evolve(c, args=((a, a_axes), (new_b, b_axes)), out_axes=out_axes)
+        c_p = Einsum((a, a_axes), (new_b, b_axes), out_axes=out_axes)
         assert out_axes == c_p.out_axes
         circuits.append(c_p)
 
-    return Add.from_unweighted_list(circuits)
+    return Add(*circuits)
 
 
 @given(einsum_circuit=st_distributed_einsum(), is_lhs=st.booleans())
@@ -534,7 +523,7 @@ def st_concat_add(draw: st.DrawFn):
     )
 
     assert isinstance(circuit, Add)
-    childrens_shapes = [child.shape for child in circuit.items.keys()]
+    childrens_shapes = [child.shape for child in circuit.children]
     adds = [circuit]
 
     concat_axis = draw(st.integers(min_value=0, max_value=len(circuit.shape) - 1))
@@ -555,9 +544,9 @@ def st_concat_add(draw: st.DrawFn):
                 )
             )
 
-        adds.append(Add({child: draw(st_c.bounded_floats) for child in children}))
+        adds.append(Add.from_weighted_nodes(*((child, draw(st_c.bounded_floats)) for child in children)))
 
-    return Concat(tuple(adds), axis=concat_axis)
+    return Concat(*adds, axis=concat_axis)
 
 
 # @hypothesis.settings(suppress_health_check=[hypothesis.HealthCheck.filter_too_much, hypothesis.HealthCheck.too_slow])
@@ -582,7 +571,7 @@ def st_add_concat(draw: st.DrawFn):
     )
 
     assert isinstance(circuit, Concat)
-    childrens_shapes = [child.shape for child in circuit.circuits]
+    childrens_shapes = [child.shape for child in circuit.children]
     concats = [circuit]
 
     sum_length = draw(st.integers(min_value=0, max_value=5))
@@ -605,13 +594,13 @@ def st_add_concat(draw: st.DrawFn):
 
         concats.append(
             Concat(
-                tuple(child for child in children),
+                *children,
                 # tuple((child, draw(st_c.bounded_floats)) for child in children),
                 axis=circuit.axis,
             )
         )
 
-    return Add.from_weighted_list([(concat, draw(st_c.bounded_floats)) for concat in concats])
+    return Add.from_weighted_nodes(*[(concat, draw(st_c.bounded_floats)) for concat in concats])
 
 
 @hypothesis.settings(
@@ -626,29 +615,22 @@ def test_pull_through_add_concat(circuit):
     circuit=st_Circuit(
         st_np.array_shapes(min_dims=0),
         must_be_explicitly_computable=True,
-        max_growth_steps=4,  # 1 + num_children of the Add.
+        max_growth_steps=10,
         growth_order="breadth_first",
         probs_default=CP.kw(all=1),
         probs_per_depth=[CP.kw(all=0, Add=(1, dict(max_n_children=3))), CP.kw(all=0, Index=1)],
         add_must_have_weight_one=True,
     )
 )
+@example(circuit=Add(Index(node=Scalar(0.0, (1,)), index=(0,)), Index(node=Scalar(1.0, (1,)), index=(0,))))
 @example(
-    circuit=Add.from_unweighted_list(
-        [Index(node=Zero(shape=(1,)), index=(0,)), Index(node=One(shape=(1,)), index=(0,))]
-    )
-)
-@example(
-    circuit=Add.from_unweighted_list(
-        [
-            Index(node=Zero(shape=(25, 30)), index=(slice(2, 12), 3)),
-            Index(node=One(shape=(25, 30)), index=(slice(2, 12), 3)),
-        ]
+    circuit=Add(
+        Index(node=Scalar(0.0, shape=(25, 30)), index=(slice(2, 12), 3)),
+        Index(node=Scalar(1.0, shape=(25, 30)), index=(slice(2, 12), 3)),
     )
 )
 def test_pull_through_add_index(circuit: Add):
-    circuit = Add.unwrap(circuit)
-    assert all(isinstance(c, Index) for c in circuit.children())
+    assert all(isinstance(c, Index) for c in circuit.children)
     check_transform_equality(pull_through_add_index, circuit)
 
 
@@ -661,7 +643,7 @@ def test_pull_through_add_index(circuit: Add):
         probs_default=CP.kw(all=1),
         probs_per_depth=[
             CP.kw(all=0, Add=1),
-            CP.kw(all=1, UnaryRearrange=(50, dict(do_squeeze=False))),
+            CP.kw(all=1, Rearrange=(50, dict(do_squeeze=False))),
         ],
         add_must_have_weight_one=True,
     )
@@ -677,37 +659,20 @@ def test_pull_through_add_rearrange(circuit):
         max_growth_steps=3,
         probs_per_depth=[
             CP.kw(all=0, Einsum=1),
-            CP.kw(
-                all=1, UnaryRearrange=(50, dict(do_split=False, do_unsqueeze=False, do_repeat=False, do_squeeze=False))
-            ),
+            CP.kw(all=1, Rearrange=(50, dict(do_split=False, do_unsqueeze=False, do_repeat=False, do_squeeze=False))),
         ],
     )
 )
 @hypothesis.example(
     circuit=Einsum(
-        args=(
-            (
-                UnaryRearrange(
-                    node=Zero(shape=(1, 2, 1), name=""),
-                    op_string="a c b -> a b c",
-                    axes_lengths=(("a", 1), ("b", 1), ("c", 2)),
-                    name="",
-                    shape=(1, 1, 2),
-                    is_constant=True,
-                    is_explicitly_computable=True,
-                    can_be_sampled=True,
-                ),
-                (0, 0, 1),
-            ),
-            (Zero(shape=(), name=""), ()),
-            (Zero(shape=(1,), name=""), (0,)),
+        (
+            rc.Rearrange.from_string(Scalar(0.0, shape=(1, 2, 1), name=""), "a:1 c:2 b:1 -> a b c"),
+            (0, 0, 1),
         ),
+        (Scalar(0.0, shape=(), name=""), ()),
+        (Scalar(0.0, shape=(1,), name=""), (0,)),
         out_axes=(0, 1),
         name="given_test",
-        shape=(1, 2),
-        is_constant=True,
-        is_explicitly_computable=True,
-        can_be_sampled=True,
     ),
 )
 def test_fuse_einsum_permute(circuit: Einsum):
@@ -721,7 +686,7 @@ def test_fuse_einsum_permute(circuit: Einsum):
         max_growth_steps=3,
         probs_per_depth=[
             CP.kw(all=0, Einsum=1),
-            CP.kw(all=1, UnaryRearrange=(50, dict(do_split=False, do_unsqueeze=True, do_repeat=True))),
+            CP.kw(all=1, Rearrange=(50, dict(do_split=False, do_unsqueeze=True, do_repeat=True))),
         ],
     )
 )
@@ -735,27 +700,16 @@ def test_fuse_einsum_rearrange(circuit: Einsum):
         must_be_explicitly_computable=True,
         max_growth_steps=5,
         probs_default=CP.kw(all=1),
-        probs_per_depth=[CP.kw(all=0, UnaryRearrange=1), CP.kw(all=1, Einsum=50)],
+        probs_per_depth=[CP.kw(all=0, Rearrange=1), CP.kw(all=1, Einsum=50)],
     )
 )
 @hypothesis.example(
-    circuit=UnaryRearrange(
-        node=Einsum(
-            args=((Zero(shape=(2, 1), name=""), (0, 1)),),
+    circuit=rc.Rearrange.from_string(
+        Einsum(
+            (Scalar(0.0, shape=(2, 1), name=""), (0, 1)),
             out_axes=(0, 1),
-            name="",
-            shape=(2, 1),
-            is_constant=True,
-            is_explicitly_computable=True,
-            can_be_sampled=True,
         ),
-        op_string="b a -> a b",
-        axes_lengths=(("a", 1), ("b", 2)),
-        name="",
-        shape=(1, 2),
-        is_constant=True,
-        is_explicitly_computable=True,
-        can_be_sampled=True,
+        "b:2 a:1 -> a b",
     ),
 )
 @mark_not_interesting_if(ImpossibleRewriteError)
@@ -770,8 +724,8 @@ def test_fuse_permute_einsum(circuit):
         max_growth_steps=5,
         probs_default=CP.kw(all=1),
         probs_per_depth=[
-            CP.kw(all=0, UnaryRearrange=1),
-            CP.kw(all=0, UnaryRearrange=1),
+            CP.kw(all=0, Rearrange=1),
+            CP.kw(all=0, Rearrange=1),
         ],
     )
 )
@@ -835,47 +789,37 @@ def test_fuse_einsum_single_simple():
 
 
 def test_split_einsum_concat():
-    a = ArrayConstant(torch.randn(3, 4, 5), name="a")
-    b = ArrayConstant(torch.randn(3, 5, 7), name="b")
-    c = ArrayConstant(torch.randn(3, 4, 2), name="c")
-    d = ArrayConstant(torch.randn(3, 1, 5), name="d")
-    e = ArrayConstant(torch.randn(3, 2, 5), name="e")
-    f = ArrayConstant(torch.randn(3, 3, 5), name="f")
+    a = Array(torch.randn(3, 4, 5), name="a")
+    b = Array(torch.randn(3, 5, 7), name="b")
+    c = Array(torch.randn(3, 4, 2), name="c")
+    d = Array(torch.randn(3, 1, 5), name="d")
+    e = Array(torch.randn(3, 2, 5), name="e")
+    f = Array(torch.randn(3, 3, 5), name="f")
 
-    evaluator = MemoizedFn(evaluate_fn(dtype=torch.float64))
+    # evaluator = MemoizedFn(evaluate_fn(dtype=torch.float64))
     for out in ["x w v y z", "x w v y", "z", "z w", "w", "y w", "y z", "y x"]:
-        for concat in [Concat((a, c), axis=-1), Concat((a, f, d, e), axis=1)]:
-            v = Einsum.from_einsum_str(f"x y z, x w v -> {out}", concat, b)
+        for concat in [Concat(a, c, axis=-1), Concat(a, f, d, e, axis=1)]:
+            v = Einsum.from_einsum_string(f"x y z, x w v -> {out}", concat, b)
 
             all_groups: Tuple[List[List[int]], ...] = ([[0]], [[1]], [[0], [1]], [[1], [0]])
-            if len(concat.circuits) > 2:
+            if len(concat.children) > 2:
                 all_groups += ([[1, 2]], [[3], [1, 2]], [[0], [2]], [[0], [3]])
             for groups in all_groups:
                 torch.testing.assert_close(
-                    evaluator(v),
-                    evaluator(
-                        rust_to_py(
-                            split_einsum_concat(py_to_rust(v).cast_einsum(), element_index=0, index_groups=groups)
-                        )
-                    ),
+                    v.evaluate(), split_einsum_concat(v, element_index=0, index_groups=groups).evaluate()
                 )
 
     for out in ["x w y z", "x w y", "z", "z w", "w", "y w", "y z", "y x"]:
-        for concat in [Concat((a, c), axis=-1), Concat((a, f, d, e), axis=1)]:
-            l = ArrayConstant(torch.randn(6, *concat.shape[1:]), name="z")
-            v = Einsum.from_einsum_str(f"x y z, w y z -> {out}", concat, l)
+        for concat in [Concat(a, c, axis=-1), Concat(a, f, d, e, axis=1)]:
+            l = Array(torch.randn(6, *concat.shape[1:]), name="z")
+            v = Einsum.from_einsum_string(f"x y z, w y z -> {out}", concat, l)
 
             all_groups = ([[0]], [[1]], [[0], [1]], [[1], [0]])
-            if len(concat.circuits) > 2:
+            if len(concat.children) > 2:
                 all_groups += ([[1, 2]], [[3], [0, 1, 2]], [[0], [2]], [[0], [3]])
             for groups in all_groups:
                 torch.testing.assert_close(
-                    evaluator(v),
-                    evaluator(
-                        rust_to_py(
-                            split_einsum_concat(py_to_rust(v).cast_einsum(), element_index=0, index_groups=groups)
-                        )
-                    ),
+                    v.evaluate(), split_einsum_concat(v, element_index=0, index_groups=groups).evaluate()
                 )
 
 
@@ -896,15 +840,12 @@ def test_split_einsum_concat():
 )
 @hypothesis.example(
     circuit=Einsum(
-        args=((Add(items={}, shape=()), ()), (Add(items={Zero(shape=(1,)): 0.0}, shape=(1,)), (0,))),
+        (Add(), ()),
+        (Add.from_weighted_nodes((Scalar(0.0, (1,)), 0.0)), (0,)),
         out_axes=(0,),
-        shape=(1,),
     ),
     element_index=0,
 )
-@mark_not_interesting_if(
-    CircuitTypeError, message="Requested type <class 'interp.circuit.computational_node.Add'> does not match"
-)
 def test_distribute(circuit: Einsum, element_index: int):
     assert len(circuit.args) > 0
     check_transform_equality(partial(distribute_old, element_index=element_index), circuit)
@@ -917,14 +858,12 @@ def test_distribute(circuit: Einsum, element_index: int):
         max_growth_steps=4,
         probs_default=CP.kw(all=1),
         probs_per_depth=[
-            CP.kw(
-                all=0, UnaryRearrange=(1, dict(do_split=False, do_unsqueeze=False, do_repeat=False, do_squeeze=False))
-            )
+            CP.kw(all=0, Rearrange=(1, dict(do_split=False, do_unsqueeze=False, do_repeat=False, do_squeeze=False)))
         ],
     )
 )
-def test_permute_to_einsum(circuit: UnaryRearrange):
-    assert circuit.get_spec().is_permute()
+def test_permute_to_einsum(circuit: Rearrange):
+    assert circuit.spec.is_permute()
     check_transform_equality(permute_to_einsum, circuit)
 
 
@@ -935,16 +874,14 @@ def test_permute_to_einsum(circuit: UnaryRearrange):
         max_growth_steps=4,
         probs_default=CP.kw(all=1),
         probs_per_depth=[
-            CP.kw(
-                all=0, UnaryRearrange=(1, dict(do_split=False, do_unsqueeze=False, do_repeat=False, do_squeeze=False))
-            ),
+            CP.kw(all=0, Rearrange=(1, dict(do_split=False, do_unsqueeze=False, do_repeat=False, do_squeeze=False))),
             CP.kw(all=0, Add=1),
         ],
     )
 )
 @mark_not_interesting_if(NotImplementedError)
-def test_push_down_permute_via_einsum(circuit: UnaryRearrange):
-    assert circuit.get_spec().is_permute()
+def test_push_down_permute_via_einsum(circuit: Rearrange):
+    assert circuit.spec.is_permute()
     assert isinstance(circuit.node, Add)
     check_transform_equality(push_down_permute_via_einsum, circuit)
 
@@ -965,7 +902,7 @@ def st_explicit_reduce_args(draw: st.DrawFn, must_be_explicitly_computable: bool
     assert isinstance(circuit, Einsum) and len(circuit.args) > 0
 
     which = draw(st.integers(min_value=0, max_value=len(circuit.args) - 1))
-    circ = circuit.get_info_by(which)[0][0]
+    circ = circuit.children[which]
     if circ.ndim == 0:
         # Attempt the rewrite anyways, should raise ImpossibleRewriteError
         return circuit, 0, which, []
@@ -1045,7 +982,7 @@ def test_explicit_reduce(explicit_reduce_args: ExplicitReduceArgs):
         probs_per_depth=[CP.kw(all=0, Einsum=1), CP.kw(all=1, Einsum=70)],
     )
 )
-@example(circuit=Einsum(args=((Einsum(args=(), out_axes=()), ()),), out_axes=()))
+@example(circuit=Einsum((Einsum(out_axes=()), ()), out_axes=()))
 @mark_not_interesting_if(
     ValueError, message="einsum(): subscript in subscript list is not within the valid range [0, 52)"
 )
@@ -1070,28 +1007,29 @@ def st_extract_add_args(draw: st.DrawFn, must_be_explicitly_computable: bool = T
     all_sub = dedup_with_order(
         draw(
             st.lists(
-                elements=st.integers(min_value=0, max_value=len(circuit.items) - 1), max_size=len(circuit.items) - 1
+                elements=st.integers(min_value=0, max_value=len(circuit.children) - 1),
+                max_size=len(circuit.children) - 1,
             )
         )
     )
 
-    circuit_items = list(circuit.items.items())
+    ch = circuit.children
 
-    return circuit, Add.from_weighted_list([circuit_items[i] for i in all_sub])
+    return circuit, Add(*(ch[i] for i in all_sub))
 
 
 @given(extract_add_args=st_extract_add_args())
 def test_extract_add(extract_add_args: ExtractAddArgs):
     circuit, sub = extract_add_args
 
-    check_transform_equality(lambda x: extract_add(x, py_to_rust(sub).cast_add()), circuit)
+    check_transform_equality(lambda x: extract_add(x, sub), circuit)
 
 
 @hypothesis.settings(deadline=None)
 @given(extract_add_args=st_extract_add_args())
 def test_extract_add_by_match(extract_add_args: ExtractAddArgs):
     circuit, sub = extract_add_args
-    matcher = Matcher(set({py_to_rust(c) for c in sub.children()}))
+    matcher = Matcher(set(sub.children))
 
     check_transform_equality(lambda x: extract_add_by_match(x, matcher), circuit)
 
@@ -1113,7 +1051,7 @@ def st_nested_einsum_axes_permute_args(
     circuit = draw(st_circuit)
     assert isinstance(circuit, Einsum)
 
-    einsums = [(i, x) for i, x in enumerate(circuit.all_input_circuits()) if isinstance(x, Einsum)]
+    einsums = [(i, x) for i, x in enumerate(circuit.children) if isinstance(x, Einsum)]
     hypothesis.assume(len(einsums) > 0, "no einsums")
     which = draw(st.integers(min_value=0, max_value=len(einsums) - 1))
     idx, einsum = einsums[which]
@@ -1137,17 +1075,17 @@ def test_nested_einsum_axes_permute(nested_einsum_axes_permute_args: NestedEinsu
 
 
 def test_nested_einsum_permute_dups_to_eq_manual():
-    x = AutoTag(Zero((2, 3)))
-    y = AutoTag(Zero((4, 5, 3)))
-    AutoTag(Zero((6, 4, 5)))
+    x = Tag.new_with_random_uuid(Scalar(0.0, (2, 3)))
+    y = Tag.new_with_random_uuid(Scalar(0.0, (4, 5, 3)))
+    Tag.new_with_random_uuid(Scalar(0.0, (6, 4, 5)))
 
-    x_y = Einsum.from_axes_tuples((x, (0, 1)), (y, (2, 3, 1)), out_axes=(2, 1, 3, 0))
-    x_y_other = Einsum.from_axes_tuples((x, (0, 1)), (y, (2, 3, 1)), out_axes=(0, 1, 3, 2))
+    x_y = Einsum((x, (0, 1)), (y, (2, 3, 1)), out_axes=(2, 1, 3, 0))
+    x_y_other = Einsum((x, (0, 1)), (y, (2, 3, 1)), out_axes=(0, 1, 3, 2))
 
-    mul = Einsum.from_axes_tuples((x_y, (2, 1, 3, 0)), (x_y_other, (7, 5, 6, 2)), out_axes=(3, 2, 6, 7))
+    mul = Einsum((x_y, (2, 1, 3, 0)), (x_y_other, (7, 5, 6, 2)), out_axes=(3, 2, 6, 7))
 
-    assert len(set(mul.all_input_circuits())) == 2
-    assert len(set(nested_einsum_permute_dups_to_eq(py_to_rust(mul).cast_einsum()).children)) == 1
+    assert len(set(mul.children)) == 2
+    assert len(set(nested_einsum_permute_dups_to_eq(mul).children)) == 1
 
 
 MultiplyAxesByIdentityArgs = Tuple[Circuit, List[int], List[int]]
