@@ -16,7 +16,6 @@ from interp.circuit.print_circuit import PrintCircuit
 from interp.circuit.testing.topdown_circuit import CircuitProbs as CP
 from interp.circuit.testing.utils import mark_not_interesting_if, rand_matcher
 from interp.tools.indexer import I
-from interp.tools.rrfs import RRFS_DIR
 from rust_circuit import (
     Add,
     Array,
@@ -77,13 +76,14 @@ from rust_circuit import (
 from rust_circuit.interop_rust import rust_to_py
 from rust_circuit.module_library import get_attention, get_bilinear_mlp, get_pointwise_mlp
 from rust_circuit.py_utils import timed
+from rust_circuit.rrfs import RRFS_DIR
 
 from .test_rust_rewrite import get_c_st
 
 P = Parser()
 
 
-@hypothesis.given(get_c_st(rust=True))
+@hypothesis.given(get_c_st())
 def test_subst_all_modules(circ):
     circ1 = rc.substitute_all_modules(circ)
     assert not any("internal_expand" in x.name for x in all_children(circ1))
@@ -92,9 +92,7 @@ def test_subst_all_modules(circ):
 
 @pytest.mark.xfail
 @hypothesis.given(
-    get_c_st(
-        rust=True, probs_default=CP.kw(all=1, Module=20, Symbol=20, Cumulant=0), from_other=False, max_growth_steps=20
-    ),
+    get_c_st(probs_default=CP.kw(all=1, Module=20, Symbol=20, Cumulant=0), from_other=False, max_growth_steps=20),
     st.data(),
 )
 @mark_not_interesting_if(rc.PushDownModulePushingPastModuleWhichOverridesSymError)
@@ -110,9 +108,7 @@ def test_push_down_modules(circ, d):
 
 @pytest.mark.xfail
 @hypothesis.given(
-    get_c_st(
-        rust=True, probs_default=CP.kw(all=1, Module=20, Symbol=20, Cumulant=0), from_other=False, max_growth_steps=5
-    ),
+    get_c_st(probs_default=CP.kw(all=1, Module=20, Symbol=20, Cumulant=0), from_other=False, max_growth_steps=5),
     st.data(),
 )
 @mark_not_interesting_if(rc.ExtractSymbolsHasBindingsFromOuterModuleError)
@@ -133,9 +129,7 @@ def test_extract_symbols(circ, d):
 
 @pytest.mark.xfail
 @hypothesis.given(
-    get_c_st(
-        rust=True, probs_default=CP.kw(all=1, Module=20, Symbol=20, Cumulant=0), from_other=False, max_growth_steps=10
-    ),
+    get_c_st(probs_default=CP.kw(all=1, Module=20, Symbol=20, Cumulant=0), from_other=False, max_growth_steps=10),
     st.data(),
 )
 def test_upd_nested_id(circ, d):
