@@ -3,12 +3,11 @@ import hypothesis.extra.numpy as st_np
 import hypothesis.strategies as st
 import pytest
 import torch
-from hypothesis import given
+from hypothesis import given, note
+from torch.testing import assert_close
 
-from interp.circuit.computational_node import Index as cIndex
-from interp.circuit.test_algebric_rewrite import check_transform_equality
+import rust_circuit as rc
 from interp.circuit.testing.topdown_circuit import CircuitProbs as CP
-from interp.circuit.testing.topdown_circuit import st_Circuit
 from interp.circuit.testing.utils import mark_not_interesting_if
 from rust_circuit import (
     Index,
@@ -23,38 +22,32 @@ from rust_circuit import (
     push_down_index_once,
     restrict,
 )
-from rust_circuit.interop_rust import py_to_rust, rust_to_py
+
+from .test_algebric_rewrites import check_transform_equality
+from .util import st_Circuit
 
 
-def run_test_push_down_index_raw(circuit: cIndex, traversal: IterativeMatcherIn = new_traversal()):
-    new_circ = rust_to_py(
-        push_down_index(py_to_rust(circuit).cast_index(), traversal=new_traversal(), allow_partial_pushdown=True)
-    )
+def run_test_push_down_index_raw(circuit: Index, options, traversal: IterativeMatcherIn = new_traversal()):
+    note(circuit.repr())
+    new_circ = push_down_index(circuit.cast_index(), traversal=traversal, **options)
     if hypothesis.currently_in_test_context():
         hypothesis.assume(new_circ != circuit)
-    check_transform_equality(
-        lambda _: new_circ,
-        circuit,
-    )
+    note(new_circ.repr())
+    x = circuit.evaluate()
+    assert_close(circuit.evaluate(), new_circ.evaluate().type(x.dtype))
 
 
 index_probs = CP.kw(
     all=0, Index=(1, dict(allow_out_of_bounds_slice=False, allow_none_slice=False, allow_neg_slice=False))
 )
 
-
-@given(
-    circuit=st_Circuit(
-        st_np.array_shapes(min_dims=0),
-        must_be_explicitly_computable=True,
-        max_growth_steps=10,
-        probs_default=CP.kw(all=1, Cumulant=0, GeneralFunction=0),
-        probs_per_depth=[index_probs],
+options_st = st.fixed_dictionaries(
+    dict(
+        allow_partial_pushdown=st.booleans(),
+        into_module_spec=st.booleans(),
+        elim_identity=st.booleans(),
     )
 )
-@mark_not_interesting_if(PushDownIndexError)
-def test_push_down_index(circuit: cIndex):
-    run_test_push_down_index_raw(circuit)
 
 
 @given(
@@ -63,12 +56,29 @@ def test_push_down_index(circuit: cIndex):
         must_be_explicitly_computable=True,
         max_growth_steps=10,
         probs_per_depth=[index_probs],
+        rust=True,
     ),
+    options=options_st,
+)
+@mark_not_interesting_if(PushDownIndexError)
+def test_push_down_index(circuit: Index, options):
+    run_test_push_down_index_raw(circuit, options)
+
+
+@given(
+    circuit=st_Circuit(
+        st_np.array_shapes(min_dims=0),
+        must_be_explicitly_computable=True,
+        max_growth_steps=10,
+        probs_per_depth=[index_probs],
+        rust=True,
+    ),
+    options=options_st,
     end_depth=st.integers(min_value=1, max_value=5),
 )
 @mark_not_interesting_if(PushDownIndexError)
-def test_push_down_index_variable_iters(circuit: cIndex, end_depth: int):
-    run_test_push_down_index_raw(circuit, restrict(default_index_traversal(), end_depth=end_depth))
+def test_push_down_index_variable_iters(circuit: Index, options, end_depth: int):
+    run_test_push_down_index_raw(circuit, options, restrict(default_index_traversal(), end_depth=end_depth))
 
 
 @given(
@@ -77,11 +87,13 @@ def test_push_down_index_variable_iters(circuit: cIndex, end_depth: int):
         must_be_explicitly_computable=True,
         max_growth_steps=4,
         probs_per_depth=[index_probs, CP.kw(all=0, Concat=1)],
-    )
+        rust=True,
+    ),
+    options=options_st,
 )
 @mark_not_interesting_if(PushDownIndexError)
-def test_push_down_index_concat(circuit: cIndex):
-    run_test_push_down_index_raw(circuit)
+def test_push_down_index_concat(circuit: Index, options):
+    run_test_push_down_index_raw(circuit, options)
 
 
 @given(
@@ -90,11 +102,13 @@ def test_push_down_index_concat(circuit: cIndex):
         must_be_explicitly_computable=True,
         max_growth_steps=4,
         probs_per_depth=[index_probs, CP.kw(all=0, Einsum=1)],
-    )
+        rust=True,
+    ),
+    options=options_st,
 )
 @mark_not_interesting_if(PushDownIndexError)
-def test_push_down_index_einsum(circuit: cIndex):
-    run_test_push_down_index_raw(circuit)
+def test_push_down_index_einsum(circuit: Index, options):
+    run_test_push_down_index_raw(circuit, options)
 
 
 # @given(
@@ -106,7 +120,7 @@ def test_push_down_index_einsum(circuit: cIndex):
 #     )
 # )
 # @mark_not_interesting_if(PushDownIndexError)
-# def test_push_down_index_cumulant(circuit: cIndex):
+# def test_push_down_index_cumulant(circuit: Index):
 #     run_test_push_down_index_raw(circuit)
 
 
