@@ -7,8 +7,6 @@ import pytest
 import torch
 
 import rust_circuit as rc
-from interp.circuit.circuit import Circuit as cCircuit
-from rust_circuit.interop_rust import py_to_rust
 from rust_circuit.py_utils import timed
 
 from .test_rust_rewrite import get_c_st
@@ -20,7 +18,7 @@ def print_multiline_escape(s: str):
 
 
 @st.composite
-def st_maybe_many_circuits(draw: st.DrawFn, *args, **kwargs) -> Tuple[List[cCircuit], bool]:
+def st_maybe_many_circuits(draw: st.DrawFn, *args, **kwargs) -> Tuple[List[rc.Circuit], bool]:
     c_st = get_c_st(*args, **kwargs)
     many = draw(st.booleans())
     if many:
@@ -37,14 +35,13 @@ def st_maybe_many_circuits(draw: st.DrawFn, *args, **kwargs) -> Tuple[List[cCirc
 @pytest.mark.parametrize("force_use_serial_numbers", [False, True])
 @pytest.mark.parametrize("only_child_below", [False, True])
 def test_rust_printing_bijection(
-    cs_is_many: Tuple[List[cCircuit], bool],
+    cs_is_many: Tuple[List[rc.Circuit], bool],
     names: bool,
     shapes_only_necessary: bool,
     force_use_serial_numbers: bool,
     only_child_below: bool,
 ):
-    cs, is_many = cs_is_many
-    circs = [py_to_rust(c) for c in cs]
+    circs, is_many = cs_is_many
     if not names:
         circs = [rc.strip_names_and_tags(circ) for circ in circs]
     printed = rc.PrintOptions(
@@ -53,12 +50,13 @@ def test_rust_printing_bijection(
         force_use_serial_numbers=force_use_serial_numbers,
         only_child_below=only_child_below,
     ).repr(*circs)
-    print(printed)
+    hypothesis.note(printed)
     # print_multiline_escape(printed)
+    parser = rc.Parser(module_check_all_inputs_used=False)
     if is_many:
-        parsed = rc.Parser().parse_circuits(printed)
+        parsed = parser.parse_circuits(printed)
     else:
-        parsed = [rc.Parser().parse_circuit(printed)]
+        parsed = [parser.parse_circuit(printed)]
     parsed = [rc.deep_canonicalize(p) for p in parsed]
     circs = [rc.deep_canonicalize(c) for c in circs]
     if parsed is not None:
@@ -699,7 +697,7 @@ def test_dtype_commenter():
   1 Array # int8"""
     )
     assert (
-        rc.PrintOptions(commenters=[rc.PrintOptions.dtype_commenter()]).repr(rc.Scalar(0.7, [1]))
+        rc.PrintOptions(commenters=[rc.PrintOptions.dtype_commenter()]).repr(rc.Scalar(0.7, (1,)))
         == """0 [1] Scalar 0.7"""
     )
 
@@ -828,7 +826,7 @@ def test_print_color():
         ),
     ]
     print()
-    for (colorer, expected) in colorers:
+    for colorer, expected in colorers:
         print()
         rc.PrintOptions(colorer=colorer).print(circ)
         print()
@@ -884,14 +882,13 @@ def test_size_colorer():
 
 
 @hypothesis.settings(deadline=None)
-@hypothesis.given(c=get_c_st())
-def test_self_hash(c: cCircuit):
+@hypothesis.given(circ=get_c_st())
+def test_self_hash(circ: rc.Circuit):
     try:
-        rust_circ = py_to_rust(c)
-        replaced = rust_circ.map_children(lambda x: rc.Scalar(1.0, x.shape, None))
-        assert rc.compute_self_hash(replaced) == rc.compute_self_hash(rust_circ)
+        replaced = circ.map_children(lambda x: rc.Scalar(1.0, x.shape, None))
+        assert rc.compute_self_hash(replaced) == rc.compute_self_hash(circ)
     except Exception as e:
-        print(c)
+        print(circ)
         raise e
 
 
