@@ -21,9 +21,7 @@ fn _rust(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     pyo3::anyhow::set_anyhow_to_py_err(Box::new(error::anyhow_to_py_err));
 
     use circuit_base::{
-        circuit_utils::{
-            cast_circuit, count_nodes, toposort_circuit, total_arrayconstant_size, total_flops,
-        },
+        circuit_utils::{count_nodes, toposort_circuit, total_arrayconstant_size, total_flops},
         flat_concat,
         named_axes::{propagate_named_axes, set_named_axes_py},
     };
@@ -41,9 +39,7 @@ fn _rust(py: Python<'_>, m: &PyModule) -> PyResult<()> {
             remove_add_few_input,
         },
         canonicalize::{canonicalize_node_py, deep_canonicalize_py},
-        circuit_manipulation::{
-            filter_nodes_py, path_get, replace_nodes_py, update_nodes_py, update_path_py,
-        },
+        circuit_manipulation::{filter_nodes_py, path_get, update_nodes_py, update_path_py},
         circuit_optimizer::{optimize_and_evaluate, optimize_and_evaluate_many},
         compiler_heuristics::deep_maybe_distribute_py,
         concat_rewrite::{
@@ -63,6 +59,7 @@ fn _rust(py: Python<'_>, m: &PyModule) -> PyResult<()> {
             py_circuit_to_schedule, py_circuit_to_schedule_many, scheduled_evaluate,
         },
     };
+    use circuit_utils::cast_circuit;
     use pyo3::{exceptions::PyValueError, types::PyTuple, PyTypeInfo};
 
     // we assume throughout the codebase that usize is 8 bytes, and otherwise error here
@@ -98,7 +95,7 @@ fn _rust(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_class::<circuit_base::generalfunction::GeneralFunctionSetDDSpec>()?;
     m.add_class::<circuit_base::generalfunction::GeneralFunctionPowSpec>()?;
     m.add_class::<circuit_base::generalfunction::GeneralFunctionMultinomialSpec>()?;
-    m.add_class::<circuit_base::generalfunction::GeneralFunctionSpecTester>()?;
+    m.add_class::<nb_operations::generalfunction::GeneralFunctionSpecTester>()?;
     m.add_class::<circuit_base::Einsum>()?;
     m.add_class::<circuit_base::Array>()?;
     m.add_class::<circuit_base::Symbol>()?;
@@ -158,7 +155,7 @@ fn _rust(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_class::<get_update_node::Expander>()?;
     m.add_class::<circuit_base::print::PrintOptions>()?;
     m.add_class::<circuit_base::print_html::PrintHtmlOptions>()?;
-    m.add_class::<circuit_base::set_of_circuits::SetOfCircuitIdentities>()?;
+    m.add_class::<circuit_utils::SetOfCircuitIdentities>()?;
     m.add_function(wrap_pyfunction!(
         circuit_base::print::set_debug_print_options,
         m
@@ -204,10 +201,13 @@ fn _rust(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_function(wrap_pyfunction!(concat_merge_uniform, m)?)?;
     m.add_function(wrap_pyfunction!(generalfunction_pull_removable_axes, m)?)?;
     m.add_function(wrap_pyfunction!(
-        circuit_base::upcast_tensor_device_dtypes_py,
+        rr_util::tensor_util::upcast_tensor_device_dtypes_py,
+        m
+    )?)?;
+    m.add_function(wrap_pyfunction!(
+        rr_util::tensor_util::upcast_tensor_devices_py,
         m
     )?)?;
-    m.add_function(wrap_pyfunction!(circuit_base::upcast_tensor_devices_py, m)?)?;
     m.add_function(wrap_pyfunction!(
         circuit_rewrites::generalfunction_rewrite::generalfunction_merge_inverses,
         m
@@ -347,7 +347,7 @@ fn _rust(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         m
     )?)?;
     m.add_function(wrap_pyfunction!(
-        circuit_base::circuit_utils::replace_all_randn_seeded,
+        circuit_utils::replace_all_randn_seeded,
         m
     )?)?;
 
@@ -368,7 +368,7 @@ fn _rust(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_function(wrap_pyfunction!(circuit_base::all_children, m)?)?;
 
     m.add_function(wrap_pyfunction!(filter_nodes_py, m)?)?;
-    m.add_function(wrap_pyfunction!(replace_nodes_py, m)?)?;
+    m.add_function(wrap_pyfunction!(circuit_utils::replace_nodes_py, m)?)?;
     m.add_function(wrap_pyfunction!(update_nodes_py, m)?)?;
     m.add_function(wrap_pyfunction!(path_get, m)?)?;
     m.add_function(wrap_pyfunction!(update_path_py, m)?)?;
@@ -562,7 +562,15 @@ fn _rust(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         nb_operations::modules::extract_symbols_get,
         m
     )?)?;
+    m.add_function(wrap_pyfunction!(
+        nb_operations::modules::pull_out_of_modules,
+        m
+    )?)?;
 
+    m.add_function(wrap_pyfunction!(
+        nb_operations::modules::pull_out_of_modules_get,
+        m
+    )?)?;
     // create dummy object for all the type aliases we use in the stub file
     // TODO: maybe these should somehow be the actual types we use in the stub file
     // (this is kinda annoying)
